@@ -3,7 +3,7 @@
  * raptor_abbrev.c - Code common to abbreviating serializers (ttl/rdfxmla)
  *
  * Copyright (C) 2006, Dave Robillard
- * Copyright (C) 2004-2006, David Beckett http://purl.org/net/dajobe/
+ * Copyright (C) 2004-2008, David Beckett http://www.dajobe.org/
  * Copyright (C) 2004-2005, University of Bristol, UK http://www.bristol.ac.uk/
  * Copyright (C) 2005, Steve Shepard steveshep@gmail.com
  * 
@@ -58,19 +58,21 @@
  **/
 
 raptor_abbrev_node* 
-raptor_new_abbrev_node(raptor_identifier_type node_type, const void *node_data,
+raptor_new_abbrev_node(raptor_world* world,
+                       raptor_identifier_type node_type, const void *node_data,
                        raptor_uri *datatype, const unsigned char *language)
 {
   unsigned char *string;
   raptor_abbrev_node* node=NULL;
   
   if(node_type == RAPTOR_IDENTIFIER_TYPE_UNKNOWN)
-    return 0;
+    return NULL;
 
   node = (raptor_abbrev_node*)RAPTOR_CALLOC(raptor_abbrev_node, 1,
                                             sizeof(raptor_abbrev_node));
 
   if(node) {
+    node->world = world;
     node->ref_count = 1;
     node->type = node_type;
     
@@ -79,12 +81,14 @@ raptor_new_abbrev_node(raptor_identifier_type node_type, const void *node_data,
           node->type = RAPTOR_IDENTIFIER_TYPE_RESOURCE;
           /* intentional fall through */
         case RAPTOR_IDENTIFIER_TYPE_RESOURCE:
-          node->value.resource.uri = raptor_uri_copy((raptor_uri*)node_data);
+          node->value.resource.uri = raptor_uri_copy_v2(world, (raptor_uri*)node_data);
           break;
           
         case RAPTOR_IDENTIFIER_TYPE_ANONYMOUS:
           string=(unsigned char*)RAPTOR_MALLOC(blank,
                                                strlen((char*)node_data)+1);
+          if(!string)
+            goto oom;
           strcpy((char*)string, (const char*) node_data);
           node->value.blank.string = string;
           break;
@@ -97,17 +101,23 @@ raptor_new_abbrev_node(raptor_identifier_type node_type, const void *node_data,
         case RAPTOR_IDENTIFIER_TYPE_XML_LITERAL:
           string = (unsigned char*)RAPTOR_MALLOC(literal,
                                                  strlen((char*)node_data)+1);
+          if(!string)
+            goto oom;
           strcpy((char*)string, (const char*)node_data);
           node->value.literal.string = string;
 
           if(datatype) {
-            node->value.literal.datatype = raptor_uri_copy(datatype);
+            node->value.literal.datatype = raptor_uri_copy_v2(world, datatype);
           }
 
           if(language) {
             unsigned char *lang;
             lang=(unsigned char*)RAPTOR_MALLOC(language,
                                                strlen((const char*)language)+1);
+            if(!lang) {
+              RAPTOR_FREE(literal, string);
+              goto oom;
+            }
             strcpy((char*)lang, (const char*)language);
             node->value.literal.language = lang;
           }
@@ -121,14 +131,18 @@ raptor_new_abbrev_node(raptor_identifier_type node_type, const void *node_data,
   }
 
   return node;
+
+  /* out of memory - clean up and return NULL */
+  oom:
+  RAPTOR_FREE(raptor_abbrev_node, node);
+  return NULL;
 }
 
 
 void
 raptor_free_abbrev_node(raptor_abbrev_node* node)
 {
-  if(!node)
-    return;
+  RAPTOR_ASSERT_OBJECT_POINTER_RETURN(node, raptor_abbrev_node);
 
   if(--node->ref_count)
     return;
@@ -136,7 +150,7 @@ raptor_free_abbrev_node(raptor_abbrev_node* node)
   switch (node->type) {
       case RAPTOR_IDENTIFIER_TYPE_RESOURCE:
       case RAPTOR_IDENTIFIER_TYPE_PREDICATE:
-        raptor_free_uri(node->value.resource.uri);
+        raptor_free_uri_v2(node->world, node->value.resource.uri);
         break;
           
       case RAPTOR_IDENTIFIER_TYPE_ANONYMOUS:
@@ -148,7 +162,7 @@ raptor_free_abbrev_node(raptor_abbrev_node* node)
         RAPTOR_FREE(literal, node->value.literal.string);
 
         if(node->value.literal.datatype)
-          raptor_free_uri(node->value.literal.datatype);
+          raptor_free_uri_v2(node->world, node->value.literal.datatype);
 
         if(node->value.literal.language)
           RAPTOR_FREE(language, node->value.literal.language);
@@ -166,24 +180,43 @@ raptor_free_abbrev_node(raptor_abbrev_node* node)
 }
 
 
+/**
+ * raptor_abbrev_node_cmp:
+ * @node1: node 1
+ * @node2: node 2
+ *
+ * INTERNAL -compare two raptor_abbrev_nodes.
+ *
+ * This needs to be a strong ordering for use by raptor_avltree.
+ * This is very performance critical, anything to make it faster is worth it.
+ *
+ * Return value: <0, 0 or 1 if @node1 less than, equal or greater
+ * than @node2 respectively
+ */
 int
-raptor_abbrev_node_equals(raptor_abbrev_node* node1, raptor_abbrev_node* node2)
+raptor_abbrev_node_cmp(raptor_abbrev_node* node1, raptor_abbrev_node* node2)
 {
   int rv = 0;  
 
-  if(node1->type != node2->type)
+  if(node1 == node2) {
     return 0;
+  } else if(node1->type < node2->type) {
+    return -1;
+  } else if(node1->type > node2->type) {
+    return 1;
+  }
 
   switch (node1->type) {
       case RAPTOR_IDENTIFIER_TYPE_RESOURCE:
       case RAPTOR_IDENTIFIER_TYPE_PREDICATE:
-        rv = raptor_uri_equals(node1->value.resource.uri,
-                               node2->value.resource.uri);
+        rv = raptor_uri_compare_v2(node1->world,
+                                   node1->value.resource.uri,
+                                   node2->value.resource.uri);
         break;
           
       case RAPTOR_IDENTIFIER_TYPE_ANONYMOUS:
-        rv = !strcmp((const char*)node1->value.blank.string,
-                     (const char*)node2->value.blank.string);
+        rv = strcmp((const char*)node1->value.blank.string,
+                    (const char*)node2->value.blank.string);
         break;
           
       case RAPTOR_IDENTIFIER_TYPE_LITERAL:
@@ -193,29 +226,43 @@ raptor_abbrev_node_equals(raptor_abbrev_node* node1, raptor_abbrev_node* node2)
             (char *)node2->value.literal.string != NULL) {
 
           /* string */
-          rv = (strcmp((char *)node1->value.literal.string,
-                       (char *)node2->value.literal.string) == 0);
+          rv = strcmp((const char*)node1->value.literal.string,
+                       (const char*)node2->value.literal.string);
+          if(rv != 0) 
+            break;
 
-          /* language */
-          if((char *)node1->value.literal.language != NULL &&
-              (char *)node2->value.literal.language != NULL) {
-            rv &= (strcmp((char *)node1->value.literal.language,
-                          (char *)node2->value.literal.language) == 0);
-          } else if((char *)node1->value.literal.language != NULL ||
-                     (char *)node2->value.literal.language != NULL) {
+          /* if strings are equal, compare language */
+          if(node1->value.literal.language == NULL &&
+             node2->value.literal.language == NULL) {
             rv = 0;
+          } else if(node1->value.literal.language != NULL &&
+                    node2->value.literal.language != NULL) {
+            rv = strcmp((const char*)node1->value.literal.language,
+                        (const char*)node2->value.literal.language);
+          } else if(node1->value.literal.language == NULL) {
+            rv = -1;
+          } else {
+            rv = 1;
           }
+          if(rv != 0) 
+            break;
 
-          /* datatype */
-          if(node1->value.literal.datatype != NULL &&
-              node2->value.literal.datatype != NULL) {
-            rv &= (raptor_uri_equals(node1->value.literal.datatype,
-                                     node2->value.literal.datatype) != 0);
-          } else if(node1->value.literal.datatype != NULL ||
-                     node2->value.literal.datatype != NULL) {
+          /* if string and language are equal, compare datatype */
+          if(node1->value.literal.datatype == NULL &&
+             node2->value.literal.datatype == NULL ) {
             rv = 0;
+          } else if(node1->value.literal.datatype != NULL &&
+                    node2->value.literal.datatype != NULL) {
+            rv = strcmp((char*)node1->value.literal.datatype,
+                  (char*)node2->value.literal.datatype);
+          } else if(node1->value.literal.datatype == NULL) {
+            rv = -1;
+          } else {
+            rv = 1;
           }
           
+          /* if rv = 0 here then the nodes are completely equal */
+
         } else {
           RAPTOR_FATAL1("string must be non-NULL for literal or xml literal\n");
           rv = 0;
@@ -224,7 +271,13 @@ raptor_abbrev_node_equals(raptor_abbrev_node* node1, raptor_abbrev_node* node2)
         break;
           
       case RAPTOR_IDENTIFIER_TYPE_ORDINAL:
-        rv = (node1->value.ordinal.ordinal == node2->value.ordinal.ordinal);
+        if(node1->value.ordinal.ordinal == node2->value.ordinal.ordinal)
+          rv = 0;
+        else if(node1->value.ordinal.ordinal < node2->value.ordinal.ordinal)
+          rv = -1;
+        else
+          rv = 1;
+
         break;
         
       case RAPTOR_IDENTIFIER_TYPE_UNKNOWN: 
@@ -237,8 +290,14 @@ raptor_abbrev_node_equals(raptor_abbrev_node* node1, raptor_abbrev_node* node2)
   
 }
 
+int
+raptor_abbrev_node_equals(raptor_abbrev_node* node1, raptor_abbrev_node* node2)
+{
+  return raptor_abbrev_node_cmp(node1, node2) == 0;
+}
 
-/*
+
+/**
  * raptor_abbrev_node_matches:
  * @node: #raptor_abbrev_node to compare
  * @node_type: Raptor identifier type
@@ -247,6 +306,9 @@ raptor_abbrev_node_equals(raptor_abbrev_node* node1, raptor_abbrev_node* node2)
  * @datatype: Literal datatype or NULL
  * @language: Literal language or NULL
  *
+ * INTERNAL - Compare an existing abbrev node against node described by
+ * parameters
+ *
  * Return value: non-zero if @node matches the node described by the rest of
  *   the parameters.
  */
@@ -264,8 +326,8 @@ raptor_abbrev_node_matches(raptor_abbrev_node* node,
   switch (node->type) {
       case RAPTOR_IDENTIFIER_TYPE_RESOURCE:
       case RAPTOR_IDENTIFIER_TYPE_PREDICATE:
-        rv = raptor_uri_equals(node->value.resource.uri,
-                               (raptor_uri *)node_data);
+        rv = raptor_uri_equals_v2(node->world, node->value.resource.uri,
+                                  (raptor_uri *)node_data);
         break;
           
       case RAPTOR_IDENTIFIER_TYPE_ANONYMOUS:
@@ -294,7 +356,7 @@ raptor_abbrev_node_matches(raptor_abbrev_node* node,
 
           /* datatype */
           if(node->value.literal.datatype != NULL && datatype != NULL)
-            rv &= (raptor_uri_equals(node->value.literal.datatype,datatype) !=0);
+            rv &= (raptor_uri_equals_v2(node->world, node->value.literal.datatype,datatype) !=0);
           else if(node->value.literal.datatype != NULL || datatype != NULL)
             rv = 0;
           
@@ -319,55 +381,129 @@ raptor_abbrev_node_matches(raptor_abbrev_node* node,
 }
 
 
-/*
+/**
  * raptor_abbrev_node_lookup:
  * @nodes: Sequence of nodes to search
  * @node_type: Raptor identifier type
- * @node_value: Node value to search with (using raptor_abbrev_node_matches).
+ * @node_value: Node value to search with using raptor_abbrev_node_matches().
  * @datatype: Literal datatype or NULL
  * @language: Literal language or NULL
+ * @created_p: (output parameter) set to non-0 if a node was created
  *
- * Return value: non-zero if @node matches the node described by the rest of
- *   the parameters.
+ * INTERNAL - Look in an avltree of nodes for a node described by parameters
+ *   and if present create it, add it and return it
+ *
+ * Return value: the node found/created or NULL on failure
  */
 raptor_abbrev_node* 
-raptor_abbrev_node_lookup(raptor_sequence* nodes,
+raptor_abbrev_node_lookup(raptor_avltree* nodes,
                           raptor_identifier_type node_type,
                           const void *node_value, raptor_uri *datatype,
-                          const unsigned char *language)
+                          const unsigned char *language, int* created_p)
 {
-  raptor_abbrev_node* rv_node = NULL;
-  int i;
-  
-  /* Search for specified node in array. TODO: this should really be a
-   * hash, not a list. */
-  for(i=0; i < raptor_sequence_size(nodes); i++) {
-    raptor_abbrev_node* node = (raptor_abbrev_node*)raptor_sequence_get_at(nodes, i);
+  raptor_abbrev_node *lookup_node;
+  raptor_abbrev_node *rv_node;
 
-    if(raptor_abbrev_node_matches(node, node_type, node_value, datatype,
-                                  language)) {
-      rv_node = node;
-      break;
-    }
-  }
+  /* Create a temporary node for search comparison. */
+  lookup_node = raptor_new_abbrev_node(nodes->world, node_type, node_value, datatype, language);
   
-  /* If not found, create one and insert it */
+  if(!lookup_node)
+    return NULL;
+
+  rv_node=(raptor_abbrev_node*)raptor_avltree_search(nodes, lookup_node);
+  
+  if(created_p)
+    *created_p=(!rv_node);
+  
+  /* If not found, insert/return a new one */
   if(!rv_node) {
-    rv_node = raptor_new_abbrev_node(node_type, node_value, datatype, language);
     
-    if(rv_node) {
-      if(raptor_sequence_push(nodes, rv_node)) {
-        raptor_free_abbrev_node(rv_node);
-        rv_node = NULL;
-      }
-      
+    if(raptor_avltree_add(nodes, lookup_node)) {
+      /* Insert failed */
+      raptor_free_abbrev_node(lookup_node);
+      return NULL;
+    } else {
+      return lookup_node;
     }
-    
+
+  /* Found */
+  } else {
+    raptor_free_abbrev_node(lookup_node);
+    return rv_node;
   }
-  
-  return rv_node;
 }
 
+
+static raptor_abbrev_node**
+raptor_new_abbrev_po(raptor_abbrev_node* predicate, raptor_abbrev_node* object)
+{
+  raptor_abbrev_node** nodes=NULL;
+  nodes = (raptor_abbrev_node**)RAPTOR_CALLOC(raptor_abbrev_nodes, 2,
+                                              sizeof(raptor_abbrev_node*));
+  if(!nodes)
+    return NULL;
+  
+  nodes[0]=predicate;
+  nodes[1]=object;
+  
+  return nodes;
+}
+
+
+static void
+raptor_free_abbrev_po(raptor_abbrev_node** nodes)
+{
+  RAPTOR_ASSERT_OBJECT_POINTER_RETURN(nodes, raptor_abbrev_node_pair);
+  
+  if(nodes[0])
+    raptor_free_abbrev_node(nodes[0]);
+  if(nodes[1])
+    raptor_free_abbrev_node(nodes[1]);
+
+  RAPTOR_FREE(raptor_abbrev_nodes, nodes);
+}
+
+
+static int
+raptor_compare_abbrev_po(raptor_abbrev_node** nodes1,
+                         raptor_abbrev_node** nodes2)
+{
+  int d;
+  d=raptor_abbrev_node_cmp(nodes1[0], nodes2[0]);
+  if(!d)
+    d=raptor_abbrev_node_cmp(nodes1[1], nodes2[1]);
+
+  return d;
+}
+
+
+#ifdef RAPTOR_DEBUG
+static void
+raptor_print_abbrev_po(FILE* handle, raptor_abbrev_node** nodes)
+{
+  raptor_abbrev_node* p = nodes[0];
+  raptor_abbrev_node* o = nodes[1];
+  
+  if(p && o) {
+    unsigned char *pred;
+    unsigned char *obj;
+
+    pred = raptor_statement_part_as_string_v2(p->world,
+                                              p->value.resource.uri, p->type,
+                                              NULL, NULL);
+    obj = raptor_statement_part_as_string_v2(o->world,
+                                             o->value.literal.string,
+                                             o->type,
+                                             o->value.literal.datatype,
+                                             o->value.literal.language);
+    fprintf(handle, "[%s : %s]\n", pred, obj);      
+    RAPTOR_FREE(cstring, pred);
+    RAPTOR_FREE(cstring, obj);
+  }
+}
+#endif
+
+
 /*
  * raptor_abbrev_subject implementation
  *
@@ -375,6 +511,8 @@ raptor_abbrev_node_lookup(raptor_sequence* nodes,
  * linked from them.
  *
  **/
+
+
 raptor_abbrev_subject*
 raptor_new_abbrev_subject(raptor_abbrev_node* node)
 {
@@ -396,8 +534,17 @@ raptor_new_abbrev_subject(raptor_abbrev_node* node)
     subject->node->count_as_subject++;
     
     subject->node_type = NULL;
+
     subject->properties =
-      raptor_new_sequence((raptor_sequence_free_handler *)raptor_free_abbrev_node, NULL);
+      raptor_new_avltree(node->world,
+                         (raptor_data_compare_function)raptor_compare_abbrev_po,
+                         (raptor_data_free_function)raptor_free_abbrev_po,
+                         0);
+#ifdef RAPTOR_DEBUG
+    raptor_avltree_set_print_handler(subject->properties,
+                                     (raptor_data_print_function)raptor_print_abbrev_po);
+#endif
+
     subject->list_items =
       raptor_new_sequence((raptor_sequence_free_handler *)raptor_free_abbrev_node, NULL);
 
@@ -415,34 +562,36 @@ raptor_new_abbrev_subject(raptor_abbrev_node* node)
 void
 raptor_free_abbrev_subject(raptor_abbrev_subject* subject) 
 {
-  if(subject) {
-    if(subject->node)
-      raptor_free_abbrev_node(subject->node);
-    
-    if(subject->node_type)
-      raptor_free_abbrev_node(subject->node_type);
-    
-    if(subject->properties)
-      raptor_free_sequence(subject->properties);
-
-    if(subject->list_items)
-      raptor_free_sequence(subject->list_items);
-
-    RAPTOR_FREE(raptor_subject, subject);
-  }
+  RAPTOR_ASSERT_OBJECT_POINTER_RETURN(subject, raptor_abbrev_subject);
   
+  if(subject->node)
+    raptor_free_abbrev_node(subject->node);
+  
+  if(subject->node_type)
+    raptor_free_abbrev_node(subject->node_type);
+  
+  if(subject->properties)
+    raptor_free_avltree(subject->properties);
+  
+  if(subject->list_items)
+    raptor_free_sequence(subject->list_items);
+  
+  RAPTOR_FREE(raptor_subject, subject);
 }
 
 
-/*
+/**
  * raptor_subject_add_property:
  * @subject: subject node to add to
  * @predicate: predicate node
  * @object: object node
+ *
+ * INTERNAL - Add predicate/object pair into properties array of a subject node.
+ *
+ * The subject node takes ownership of the predicate/object nodes.
+ * On error, predicate/object are freed immediately.
  * 
- * Add predicate/object pair into properties array of a subject node.
- * 
- * Return value: non-0 on failure
+ * Return value: <0 on failure, >0 if pair is a duplicate and it was not added
  **/
 int
 raptor_abbrev_subject_add_property(raptor_abbrev_subject* subject,
@@ -450,20 +599,43 @@ raptor_abbrev_subject_add_property(raptor_abbrev_subject* subject,
                                    raptor_abbrev_node* object) 
 {
   int err;
+  raptor_abbrev_node** nodes;
   
-  err = raptor_sequence_push(subject->properties, predicate);
-  if(err)
-    return err;
-  
-  err = raptor_sequence_push(subject->properties, object);
-  if(err) {
-    raptor_sequence_pop(subject->properties);
-    return err;
-  }
-  
+  nodes=raptor_new_abbrev_po(predicate, object);
+  if(!nodes)
+    return -1;
+
   predicate->ref_count++;
   object->ref_count++;
+
+  if(raptor_avltree_search(subject->properties, nodes)) {
+    /* Already present - do not add a duplicate triple (s->[p o]) */
+    raptor_free_abbrev_po(nodes);
+    return 1;
+  }
   
+#if 0
+  fprintf(stderr, "Adding P,O ");
+  raptor_print_abbrev_po(stderr, nodes);
+
+  raptor_avltree_dump(subject->properties, stderr);
+#endif
+  err = raptor_avltree_add(subject->properties, nodes);
+  if(err) {
+    raptor_free_abbrev_po(nodes);
+    return -1;
+  }
+#if 0
+  fprintf(stderr, "Result ");
+  raptor_avltree_print(subject->properties, stderr);
+  
+  raptor_avltree_dump(subject->properties, stderr);
+
+  raptor_avltree_check(subject->properties);
+
+  fprintf(stderr, "\n\n");
+#endif
+
   return 0;
 }
 
@@ -474,9 +646,9 @@ raptor_abbrev_subject_add_property(raptor_abbrev_subject* subject,
  * @ordinal: ordinal index
  * @object: object node
  * 
- * Add rdf:li into list element array of a #raptor_abbrev_subject node.
+ * INTERNAL - Add rdf:li into list element array of a #raptor_abbrev_subject node.
  * 
- * Return value: 
+ * Return value: non-0 on failure
  **/
 int
 raptor_abbrev_subject_add_list_element(raptor_abbrev_subject* subject, 
@@ -528,11 +700,12 @@ raptor_abbrev_subject_find(raptor_sequence *sequence,
 
 
 raptor_abbrev_subject* 
-raptor_abbrev_subject_lookup(raptor_sequence* nodes,
+raptor_abbrev_subject_lookup(raptor_avltree* nodes,
                              raptor_sequence* subjects,
                              raptor_sequence* blanks,
                              raptor_identifier_type node_type,
-                             const void *node_data)
+                             const void *node_data,
+                             int* created_p)
 {
   raptor_sequence *sequence;
   raptor_abbrev_subject* rv_subject;
@@ -545,15 +718,18 @@ raptor_abbrev_subject_lookup(raptor_sequence* nodes,
   rv_subject= raptor_abbrev_subject_find(sequence, node_type,
                                          node_data, NULL);
 
+  if(created_p)
+    *created_p=(!rv_subject);
+  
   /* If not found, create one and insert it */
   if(!rv_subject) {
     raptor_abbrev_node* node = raptor_abbrev_node_lookup(nodes, node_type,
-                                                         node_data, NULL, NULL);
+                                                         node_data, NULL, NULL,
+                                                         NULL);
     if(node) {      
       rv_subject = raptor_new_abbrev_subject(node);
       if(rv_subject) {
         if(raptor_sequence_push(sequence, rv_subject)) {
-          raptor_free_abbrev_subject(rv_subject);
           rv_subject = NULL;
         }      
       }
@@ -572,6 +748,7 @@ raptor_print_subject(raptor_abbrev_subject* subject)
   unsigned char *subj;
   unsigned char *pred;
   unsigned char *obj;
+  raptor_avltree_iterator* iter=NULL;
 
   /* Note: The raptor_abbrev_node field passed as the first argument for
    * raptor_statement_part_as_string() is somewhat arbitrary, since as
@@ -605,25 +782,20 @@ raptor_print_subject(raptor_abbrev_subject* subject)
     
   }
 
-  i=0;
-  while (i < raptor_sequence_size(subject->properties)) {
 
-    raptor_abbrev_node* p = raptor_sequence_get_at(subject->properties, i++);
-    raptor_abbrev_node* o = raptor_sequence_get_at(subject->properties, i++);
+  iter=raptor_new_avltree_iterator(subject->properties, NULL, NULL, 1);
+  while(iter) {
+    raptor_abbrev_node** nodes;
+    nodes=(raptor_abbrev_node**)raptor_avltree_iterator_get(iter);
+    if(!nodes)
+      break;
+    raptor_print_abbrev_po(stderr, nodes);
 
-    if(p && o) {
-      pred = raptor_statement_part_as_string(p->value.resource.uri, p->type,
-                                             NULL, NULL);
-      obj = raptor_statement_part_as_string(o->value.literal.string,
-                                            o->type,
-                                            o->value.literal.datatype,
-                                            o->value.literal.language);
-      fprintf(stderr,"[%s, %s, %s]\n", subj, pred, obj);      
-      RAPTOR_FREE(cstring, pred);
-      RAPTOR_FREE(cstring, obj);
-    }
-    
+    if(raptor_avltree_iterator_next(iter))
+      break;
   }
+  if(iter)
+    raptor_free_avltree_iterator(iter);
   
   RAPTOR_FREE(cstring, subj);
   
@@ -633,11 +805,11 @@ raptor_print_subject(raptor_abbrev_subject* subject)
 
 /* helper functions */
 
-/*
+/**
  * raptor_unique_id:
  * @base: base ID
  * 
- * Generate a node ID for serializing
+ * INTERNAL - Generate a node ID for serializing
  *
  * Raptor doesn't check that blank IDs it generates are unique from
  * any specified by rdf:nodeID. Here, we need to emit IDs that are
@@ -665,14 +837,14 @@ raptor_unique_id(unsigned char *base)
 }
 
 
-/*
+/**
  * raptor_new_qname_from_resource:
  * @namespaces: sequence of namespaces (corresponding to nstack)
  * @nstack: #raptor_namespace_stack to use/update
  * @namespace_count: size of nstack (may be modified)
  * @node: #raptor_abbrev_node to use 
  * 
- * Make an XML QName from the URI associated with the node.
+ * INTERNAL - Make an XML QName from the URI associated with the node.
  * 
  * Return value: the QName or NULL on failure
  **/
@@ -702,7 +874,7 @@ raptor_new_qname_from_resource(raptor_sequence* namespaces,
   if(qname)
     return qname;
   
-  uri_string = raptor_uri_as_counted_string(node->value.resource.uri, &uri_len);
+  uri_string = raptor_uri_as_counted_string_v2(node->world, node->value.resource.uri, &uri_len);
 
   p= uri_string;
   name_len=uri_len;
@@ -718,28 +890,37 @@ raptor_new_qname_from_resource(raptor_sequence* namespaces,
     return NULL;
 
   c=*name; *name='\0';
-  ns_uri=raptor_new_uri(uri_string);
+  ns_uri=raptor_new_uri_v2(node->world, uri_string);
+  if(!ns_uri)
+    return NULL;
+  
   *name=c;
   
   ns = raptor_namespaces_find_namespace_by_uri(nstack, ns_uri);
   if(!ns) {
     /* The namespace was not declared, so create one */
     unsigned char prefix[2 + MAX_ASCII_INT_SIZE + 1];
-	*namespace_count = *namespace_count + 1;
+      *namespace_count = *namespace_count + 1;
     sprintf((char *)prefix, "ns%d", *namespace_count);
 
     ns = raptor_new_namespace_from_uri(nstack, prefix, ns_uri, 0);
 
     /* We'll most likely need this namespace again. Push it on our
-     * stack.  It will be deleted in
-     * raptor_rdfxmla_serialize_terminate
+     * stack.  It will be deleted in raptor_rdfxmla_serialize_terminate()
      */
-    raptor_sequence_push(namespaces, ns);
+    if(raptor_sequence_push(namespaces, ns)) {
+      /* namespaces sequence has no free handler so we have to free
+       * the ns ourselves on error
+       */
+      raptor_free_namespace(ns);
+      raptor_free_uri_v2(node->world, ns_uri);
+      return NULL;
+    }
   }
 
-  qname = raptor_new_qname_from_namespace_local_name(ns, name,  NULL);
+  qname = raptor_new_qname_from_namespace_local_name_v2(node->world, ns, name,  NULL);
   
-  raptor_free_uri(ns_uri);
+  raptor_free_uri_v2(node->world, ns_uri);
 
   return qname;
 }
