@@ -2,7 +2,7 @@
  *
  * raptor_internal.h - Redland Parser Toolkit for RDF (Raptor) internals
  *
- * Copyright (C) 2002-2006, David Beckett http://purl.org/net/dajobe/
+ * Copyright (C) 2002-2009, David Beckett http://www.dajobe.org/
  * Copyright (C) 2002-2004, University of Bristol, UK http://www.bristol.ac.uk/
  * 
  * This package is Free Software and part of Redland http://librdf.org/
@@ -29,16 +29,20 @@
 
 #ifdef __cplusplus
 extern "C" {
+#define RAPTOR_EXTERN_C extern "C"
+#else
+#define RAPTOR_EXTERN_C
 #endif
 
 #ifdef RAPTOR_INTERNAL
 
 /* for the memory allocation functions */
-#if defined(HAVE_DMALLOC_H) && defined(RAPTOR_MEMORY_DEBUG_DMALLOC)
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #undef HAVE_STDLIB_H
 #endif
+
+#if defined(HAVE_DMALLOC_H) && defined(RAPTOR_MEMORY_DEBUG_DMALLOC)
 #include <dmalloc.h>
 #endif
 
@@ -91,7 +95,9 @@ void raptor_system_free(void *ptr);
 #define SYSTEM_FREE(ptr)   free(ptr)
 #endif
 
+#ifndef RAPTOR_ASSERT_DIE
 #define RAPTOR_ASSERT_DIE abort();
+#endif
 
 #else
 /* DEBUGGING TURNED OFF */
@@ -107,7 +113,9 @@ void raptor_system_free(void *ptr);
 #define SYSTEM_MALLOC(size)   malloc(size)
 #define SYSTEM_FREE(ptr)   free(ptr)
 
+#ifndef RAPTOR_ASSERT_DIE
 #define RAPTOR_ASSERT_DIE
+#endif
 
 #endif
 
@@ -123,7 +131,10 @@ void raptor_system_free(void *ptr);
 
 #define RAPTOR_ASSERT(condition, msg) 
 #define RAPTOR_ASSERT_RETURN(condition, msg, ret) 
-#define RAPTOR_ASSERT_OBJECT_POINTER_RETURN(pointer, type)
+#define RAPTOR_ASSERT_OBJECT_POINTER_RETURN(pointer, type) do { \
+  if(!pointer) \
+    return; \
+} while(0)
 #define RAPTOR_ASSERT_OBJECT_POINTER_RETURN_VALUE(pointer, type, ret)
 
 #else
@@ -180,7 +191,7 @@ void raptor_system_free(void *ptr);
 #endif
 
 
-typedef struct raptor_sax2_s raptor_sax2;
+
 
 #ifdef RAPTOR_XML_LIBXML
 
@@ -193,11 +204,12 @@ typedef struct raptor_sax2_s raptor_sax2;
 /* raptor_libxml.c exports */
 extern void raptor_libxml_init(raptor_sax2* sax2, raptor_uri *base_uri);
 extern void raptor_libxml_init_sax_error_handlers(xmlSAXHandler *sax);
-extern void raptor_libxml_init_generic_error_handlers(raptor_sax2 *sax2);
+extern void raptor_libxml_generic_error(void* user_data, const char *msg, ...) RAPTOR_PRINTF_FORMAT(2, 3);
 
 extern void raptor_libxml_validation_error(void *context, const char *msg, ...) RAPTOR_PRINTF_FORMAT(2, 3);
 extern void raptor_libxml_validation_warning(void *context, const char *msg, ...) RAPTOR_PRINTF_FORMAT(2, 3);
 void raptor_libxml_free(xmlParserCtxtPtr xc);
+void raptor_libxml_xmlStructuredErrorFunc(void *user_data, xmlErrorPtr err);
 
 /* raptor_parse.c - exported to libxml part */
 extern void raptor_libxml_update_document_locator(raptor_sax2* sax2, raptor_locator* locator);
@@ -226,12 +238,55 @@ typedef struct raptor_serializer_factory_s raptor_serializer_factory;
 typedef struct raptor_id_set_s raptor_id_set;
 typedef struct raptor_uri_detail_s raptor_uri_detail;
 
+/* raptor_avltree.c */
+/* AVL tree */
+typedef struct raptor_avltree_s raptor_avltree;
+typedef struct raptor_avltree_iterator_s raptor_avltree_iterator;
+
+typedef struct raptor_avltree_node_s raptor_avltree_node;
+
+/* user functions */
+typedef int (*raptor_data_compare_function)(const void* data1, const void* data2);
+typedef void (*raptor_data_free_function)(void* data);
+typedef int (*raptor_avltree_visit_function)(int depth, void* data, void *user_data);
+typedef void (*raptor_data_print_function)(FILE* handle, const void* data);
+
+/* AVL-tree */
+struct raptor_avltree_s {
+  /* raptor_world object */
+  raptor_world* world;
+
+  /* root node of tree */
+  raptor_avltree_node* root;
+
+  /* node comparison function (optional) */
+  raptor_data_compare_function compare_fn;
+
+  /* node deletion function (optional) */
+  raptor_data_free_function free_fn;
+
+  /* node print function (optional) */
+  raptor_data_print_function print_fn;
+
+  /* tree flags (none defined at present) */
+  unsigned int flags;
+
+  /* number of nodes in tree */
+  unsigned int size;
+
+  /* legacy iterator used for cursor methods */
+  raptor_avltree_iterator* cursor_iterator;
+};
+
 
 /* Raptor Namespace Stack node */
 struct raptor_namespace_stack_s {
-  raptor_namespace* top;
-  raptor_uri_handler *uri_handler;
-  void *uri_context;
+  raptor_world* world;
+  int size;
+
+  int table_size;
+  raptor_namespace** table;
+  raptor_namespace* def_namespace;
   raptor_simple_message_handler error_handler;
   void *error_data;
 
@@ -269,14 +324,19 @@ struct raptor_namespace_s {
   int is_rdf_schema;
 };
 
+raptor_namespace** raptor_namespace_stack_to_array(raptor_namespace_stack *nstack, size_t *size_p);
+
 #ifdef RAPTOR_XML_LIBXML
 #define RAPTOR_LIBXML_MAGIC 0x8AF108
 #endif
 
+
 /*
  * Raptor parser object
  */
 struct raptor_parser_s {
+  raptor_world* world;
+
 #ifdef RAPTOR_XML_LIBXML
   int magic;
 #endif
@@ -302,19 +362,15 @@ struct raptor_parser_s {
   /* stuff for our user */
   void *user_data;
 
-  void *fatal_error_user_data;
-  void *error_user_data;
-  void *warning_user_data;
-
-  raptor_message_handler fatal_error_handler;
-  raptor_message_handler error_handler;
-  raptor_message_handler warning_handler;
+  raptor_error_handlers error_handlers;
 
   void* unused1; /* UNUSED - re-use struct slot later needed */
 
   /* parser callbacks */
   raptor_statement_handler statement_handler;
 
+  raptor_graph_handler graph_handler;
+
   void *generate_id_handler_user_data;
   raptor_generate_id_handler generate_id_handler;
 
@@ -334,6 +390,30 @@ struct raptor_parser_s {
   raptor_namespace_handler namespace_handler;
 
   void* namespace_handler_user_data;
+
+  raptor_stringbuffer* sb;
+
+  /* raptor_www pointer stored here to allow cleanup on error */
+  raptor_www* www;
+
+  /* internal data for lexers */
+  void* lexer_user_data;
+
+  /* FEATURE:
+   * HTTP Cache-Control: header value to send (default NULL)
+   * RAPTOR_FEATURE_WWW_HTTP_CACHE_CONTROL
+   */
+  const char* cache_control;
+
+  /* FEATURE:
+   * HTTP User-Agent: header value to send (default NULL)
+   * RAPTOR_FEATURE_WWW_HTTP_USER_AGENT
+   */
+  const char* user_agent;
+
+  /* NOTE: if anything a user can set is added here check that
+   * raptor_parser_copy_user_state() is updated to copy it.
+   */
 };
 
 
@@ -348,6 +428,8 @@ typedef struct raptor_type_q_s raptor_type_q;
 
 /** A Parser Factory for a syntax */
 struct raptor_parser_factory_s {
+  raptor_world* world;
+
   struct raptor_parser_factory_s* next;
 
   /* syntax name */
@@ -397,6 +479,12 @@ struct raptor_parser_factory_s {
 
   /* non-0 if this parser needs a base URI */
   int need_base_uri;
+
+  /* get the current generated ID base (OPTIONAL) */
+  int (*get_current_base_id)(raptor_parser* rdf_parser);
+
+  /* get the name (OPTIONAL) */
+  const char* (*get_name)(raptor_parser* rdf_parser);
 };
 
 
@@ -404,8 +492,15 @@ struct raptor_parser_factory_s {
  * Raptor serializer object
  */
 struct raptor_serializer_s {
+  raptor_world* world;
+
   /* can be filled with error location information */
   raptor_locator locator;
+  
+  /* FEATURE:
+   * non 0 to write base URI to document header (@base)
+   */
+  int feature_write_base_uri;
 
   /* FEATURE:
    * non 0 to write relative URIs wherever possible
@@ -448,6 +543,9 @@ struct raptor_serializer_s {
 
   /* destination stream for the serialization */
   raptor_iostream *iostream;
+
+  /* if true, iostream was made here so free it */
+  int free_iostream_on_end;
   
   struct raptor_serializer_factory_s* factory;
 
@@ -458,11 +556,38 @@ struct raptor_serializer_s {
    * non 0 to write XML 1.0 or 1.1 declaration (default 1)
    */
   int feature_write_xml_declaration;
+
+  /* FEATURE:
+   * JSON serializer callback function name
+   */
+  unsigned char *feature_json_callback;
+
+  /* FEATURE:
+   * JSON serializer extra data
+   */
+  unsigned char *feature_json_extra_data;
+
+  /* FEATURE:
+   * RSS serializer triples
+   */
+  unsigned char *feature_rss_triples;
+
+  /* FEATURE:
+   * Atom serializer entry URI string
+   */
+  unsigned char *feature_atom_entry_uri;
+
+  /* FEATURE:
+   * Namespace-prefix elements OR prefer unprefixed elements.
+   */
+  int feature_prefix_elements;
 };
 
 
 /** A Serializer Factory for a syntax */
 struct raptor_serializer_factory_s {
+  raptor_world* world;
+
   struct raptor_serializer_factory_s* next;
 
   /* syntax name */
@@ -507,16 +632,28 @@ struct raptor_serializer_factory_s {
 };
 
 
+/* for raptor_parse_uri_write_bytes() when used as a handler for
+ * raptor_www_set_write_bytes_handler()
+ */
+typedef struct 
+{
+  raptor_parser* rdf_parser;
+  raptor_uri* base_uri;
+  raptor_uri* final_uri;
+  int started;
+} raptor_parse_bytes_context;
+
+
 /* raptor_serialize.c */
-void raptor_serializer_register_factory(const char *name, const char *label, const char *mime_type, const char *alias, const unsigned char *uri_string, void (*factory) (raptor_serializer_factory*));
+int raptor_serializer_register_factory(raptor_world* world, const char *name, const char *label, const char *mime_type, const char *alias, const unsigned char *uri_string, int (*factory) (raptor_serializer_factory*));
 
 
 /* raptor_general.c */
 
-raptor_parser_factory* raptor_parser_register_factory(const char *name, const char *label, void (*factory) (raptor_parser_factory*));
-void raptor_parser_factory_add_alias(raptor_parser_factory* factory, const char *alias);
-void raptor_parser_factory_add_mime_type(raptor_parser_factory* factory, const char* mime_type, int q);
-void raptor_parser_factory_add_uri(raptor_parser_factory* factory, const unsigned char *uri_string);
+raptor_parser_factory* raptor_parser_register_factory(raptor_world* world, const char *name, const char *label, int (*factory) (raptor_parser_factory*));
+int raptor_parser_factory_add_alias(raptor_parser_factory* factory, const char *alias);
+int raptor_parser_factory_add_mime_type(raptor_parser_factory* factory, const char* mime_type, int q);
+int raptor_parser_factory_add_uri(raptor_parser_factory* factory, const unsigned char *uri_string);
 
 unsigned char* raptor_parser_internal_generate_id(raptor_parser *rdf_parser, raptor_genid_type type, unsigned char *user_bnodeid);
 
@@ -524,30 +661,31 @@ unsigned char* raptor_parser_internal_generate_id(raptor_parser *rdf_parser, rap
 void raptor_stats_print(raptor_parser *rdf_parser, FILE *stream);
 #endif
 const char* raptor_basename(const char *name);
-raptor_statement* raptor_statement_copy(const raptor_statement *statement);
-void raptor_free_statement(raptor_statement *statement);
+raptor_statement* raptor_statement_copy(raptor_world* world, const raptor_statement *statement);
+raptor_statement_v2* raptor_statement_copy_v2(const raptor_statement_v2 *statement);
+raptor_statement_v2* raptor_statement_copy_v2_from_v1(raptor_world* world, const raptor_statement *statement);
+void raptor_free_statement(raptor_world* world, raptor_statement *statement);
+void raptor_free_statement_v2(raptor_statement_v2 *statement);
+void raptor_print_statement_v1(raptor_world* world, const raptor_statement * statement, FILE *stream);
 
 /* raptor_parse.c */
-raptor_parser_factory* raptor_get_parser_factory(const char *name);  
+raptor_parser_factory* raptor_get_parser_factory(raptor_world* world, const char *name);  
 void raptor_delete_parser_factories(void);
-void raptor_free_type_q(raptor_type_q* type_q);
-const char* raptor_parser_get_accept_header_all(void);
+const char* raptor_parser_get_accept_header_all(raptor_world* world);
 int raptor_parse_uri_no_net_filter(void *user_data, raptor_uri* uri);
+void raptor_parse_uri_write_bytes(raptor_www* www, void *userdata, const void *ptr, size_t size, size_t nmemb);
+void raptor_parser_fatal_error(raptor_parser* parser, const char *message, ...) RAPTOR_PRINTF_FORMAT(2, 3);
+void raptor_parser_error(raptor_parser* parser, const char *message, ...) RAPTOR_PRINTF_FORMAT(2, 3);
+void raptor_parser_simple_error(void* parser, const char *message, ...) RAPTOR_PRINTF_FORMAT(2, 3);
+void raptor_parser_error_varargs(raptor_parser* parser, const char *message, va_list arguments) RAPTOR_PRINTF_FORMAT(2, 0);
+void raptor_parser_warning(raptor_parser* parser, const char *message, ...) RAPTOR_PRINTF_FORMAT(2, 3);
 
-/* raptor_general.c */
-void raptor_invoke_message_varargs(const char *type, raptor_message_handler handler, void* user_data, raptor_locator* locator, const char *message, va_list arguments) RAPTOR_PRINTF_FORMAT(5, 0);
-void raptor_invoke_simple_message_varargs(const char *type, raptor_simple_message_handler handler, void* user_data, const char *message, va_list arguments) RAPTOR_PRINTF_FORMAT(4, 0);
+/* logging */
+#define RAPTOR_ERROR_HANDLER_MAGIC 0xD00DB1FF
 
-extern void raptor_parser_fatal_error(raptor_parser* parser, const char *message, ...) RAPTOR_PRINTF_FORMAT(2, 3);
-extern void raptor_parser_error(raptor_parser* parser, const char *message, ...) RAPTOR_PRINTF_FORMAT(2, 3);
-extern void raptor_parser_simple_error(void* parser, const char *message, ...) RAPTOR_PRINTF_FORMAT(2, 3);
-extern void raptor_parser_error_message_handler(void *user_data, raptor_locator* locator, const char *message);
-extern void raptor_parser_warning(raptor_parser* parser, const char *message, ...) RAPTOR_PRINTF_FORMAT(2, 3);
-extern void raptor_parser_fatal_error_varargs(raptor_parser* parser, const char *message, va_list arguments) RAPTOR_PRINTF_FORMAT(2, 0);
-extern void raptor_parser_fatal_error_message_handler(void *user_data, raptor_locator* locator, const char *message);
-extern void raptor_parser_error_varargs(raptor_parser* parser, const char *message, va_list arguments) RAPTOR_PRINTF_FORMAT(2, 0);
-extern void raptor_parser_warning_varargs(raptor_parser* parser, const char *message, va_list arguments)  RAPTOR_PRINTF_FORMAT(2, 0);
-void raptor_parser_warning_message_handler(void *user_data, raptor_locator* locator, const char *message);
+void raptor_log_error_to_handlers(raptor_world* world, raptor_error_handlers* error_handlers, raptor_log_level level, raptor_locator* locator, const char* message);
+void raptor_log_error_varargs(raptor_world* world, raptor_log_level level, raptor_message_handler handler, void* handler_data, raptor_locator* locator, const char* message, va_list arguments) RAPTOR_PRINTF_FORMAT(6, 0);
+void raptor_log_error(raptor_world* world, raptor_log_level level, raptor_message_handler handler, void* handler_data, raptor_locator* locator, const char* message);
 
 
 /* raptor_parse.c */
@@ -562,20 +700,22 @@ extern void raptor_xml_characters_handler(void *user_data, const unsigned char *
 extern void raptor_xml_cdata_handler(void *user_data, const unsigned char *s, int len);
 void raptor_xml_comment_handler(void *user_data, const unsigned char *s);
 
-#ifdef RAPTOR_DEBUG
+#if RAPTOR_DEBUG > 1
 void raptor_rdfxml_parser_stats_print(raptor_rdfxml_parser* rdf_xml_parser, FILE *stream);
 #endif
 
-void raptor_parser_copy_user_state(raptor_parser *to_parser, raptor_parser *from_parser);
+int raptor_parser_copy_user_state(raptor_parser *to_parser, raptor_parser *from_parser);
 
 /* raptor_feature.c */
-int raptor_features_enumerate_common(const raptor_feature feature, const char **name, raptor_uri **uri, const char **label, int flags);
+int raptor_features_enumerate_common(raptor_world* world, const raptor_feature feature, const char **name, raptor_uri **uri, const char **label, int flags);
 
 /* raptor_general.c */
 extern int raptor_valid_xml_ID(raptor_parser *rdf_parser, const unsigned char *string);
 int raptor_check_ordinal(const unsigned char *name);
 
 /* raptor_identifier.c */
+void raptor_set_identifier_uri(raptor_identifier *identifier, raptor_uri *uri);
+void raptor_set_identifier_id(raptor_identifier *identifier, const unsigned char *id);
 #ifdef RAPTOR_DEBUG
 void raptor_identifier_print(FILE *stream, raptor_identifier* identifier);
 #endif
@@ -614,6 +754,7 @@ void raptor_parser_start_namespace(raptor_parser* rdf_parser, raptor_namespace*
  * only then is uri also given.
  */
 struct raptor_qname_s {
+  raptor_world* world;
   /* Name - always present */
   const unsigned char *local_name;
   int local_name_length;
@@ -635,28 +776,35 @@ void raptor_qname_print(FILE *stream, raptor_qname* name);
 
 
 /* raptor_uri.c */
-void raptor_uri_init(void);
-raptor_uri* raptor_new_uri_from_rdf_ordinal(int ordinal);
+int raptor_uri_init(raptor_world* world);
+raptor_uri* raptor_new_uri_from_rdf_ordinal(raptor_world* world, int ordinal);
 
 /* parsers */
-void raptor_init_parser_rdfxml(void);
-void raptor_init_parser_ntriples(void);
-void raptor_init_parser_turtle(void);
-void raptor_init_parser_n3(void);
-void raptor_init_parser_grddl(void);
-void raptor_init_parser_guess(void);
-void raptor_init_parser_rss(void);
+int raptor_init_parser_rdfxml(raptor_world* world);
+int raptor_init_parser_ntriples(raptor_world* world);
+int raptor_init_parser_turtle(raptor_world* world);
+int raptor_init_parser_trig(raptor_world* world);
+int raptor_init_parser_n3(raptor_world* world);
+int raptor_init_parser_grddl_common(raptor_world* world);
+int raptor_init_parser_grddl(raptor_world* world);
+int raptor_init_parser_guess(raptor_world* world);
+int raptor_init_parser_rss(raptor_world* world);
+int raptor_init_parser_rdfa(raptor_world* world);
 
-void raptor_terminate_parser_grddl(void);
+void raptor_terminate_parser_grddl_common(raptor_world *world);
 
 /* raptor_parse.c */
-void raptor_parsers_init(void);
-void raptor_parsers_finish(void);
+int raptor_parsers_init(raptor_world* world);
+void raptor_parsers_finish(raptor_world *world);
 
+void raptor_parser_save_content(raptor_parser* rdf_parser, int save);
+const unsigned char* raptor_parser_get_content(raptor_parser* rdf_parser, size_t* length_p);
+void raptor_parser_set_graph_name(raptor_parser* parser, raptor_uri* uri);
+int raptor_parser_get_current_base_id(raptor_parser* parser);
 
 /* raptor_rss.c */
-void raptor_init_serializer_rss10(void);
-void raptor_init_serializer_atom(void);
+int raptor_init_serializer_rss10(raptor_world* world);
+int raptor_init_serializer_atom(raptor_world* world);
 
 extern const unsigned char * const raptor_atom_namespace_uri;
 
@@ -666,15 +814,14 @@ void raptor_free_uri_detail(raptor_uri_detail* uri_detail);
 unsigned char* raptor_uri_detail_to_string(raptor_uri_detail *ud, size_t* len_p);
 
 /* serializers */
-void raptor_init_serializer_rdfxml(void);
-void raptor_init_serializer_ntriples(void);
-void raptor_init_serializer_dot(void);
-void raptor_init_serializer_simple(void);
-void raptor_delete_serializer_factories(void);
+int raptor_init_serializer_rdfxml(raptor_world* world);
+int raptor_init_serializer_ntriples(raptor_world* world);
+int raptor_init_serializer_dot(raptor_world* world);
+int raptor_init_serializer_simple(raptor_world* world);
 
 /* raptor_serializer.c */
-void raptor_serializers_init(void);
-void raptor_serializers_finish(void);
+int raptor_serializers_init(raptor_world* world);
+void raptor_serializers_finish(raptor_world* world);
 
 void raptor_serializer_error(raptor_serializer* serializer, const char *message, ...) RAPTOR_PRINTF_FORMAT(2, 3);
 void raptor_serializer_simple_error(void* serializer, const char *message, ...) RAPTOR_PRINTF_FORMAT(2, 3);
@@ -683,14 +830,17 @@ void raptor_serializer_warning(raptor_serializer* serializer, const char *messag
 void raptor_serializer_warning_varargs(raptor_serializer* serializer, const char *message, va_list arguments) RAPTOR_PRINTF_FORMAT(2, 0);
 
 /* raptor_serialize_rdfxmla.c */  
-void raptor_init_serializer_rdfxmla(void);
+int raptor_init_serializer_rdfxmla(raptor_world* world);
 
 /* raptor_serialize_turtle.c */  
-void raptor_init_serializer_turtle(void);
-  
+int raptor_init_serializer_turtle(raptor_world* world);
+
+/* raptor_serialize_json.c */  
+int raptor_init_serializer_json(raptor_world* world);
+
 /* raptor_utf8.c */
-int raptor_unicode_is_namestartchar(long c);
-int raptor_unicode_is_namechar(long c);
+int raptor_unicode_is_namestartchar(raptor_unichar c);
+int raptor_unicode_is_namechar(raptor_unichar c);
 int raptor_utf8_is_nfc(const unsigned char *input, size_t length);
 
 /* raptor_www*.c */
@@ -707,10 +857,13 @@ int raptor_utf8_is_nfc(const unsigned char *input, size_t length);
 #endif
 
 /* Size of buffer used in various raptor_www places for I/O  */
+#ifndef RAPTOR_WWW_BUFFER_SIZE
 #define RAPTOR_WWW_BUFFER_SIZE 4096
+#endif
 
 /* WWW library state */
 struct  raptor_www_s {
+  raptor_world* world;
   char *type;
   int free_type;
   int total_bytes;
@@ -723,13 +876,14 @@ struct  raptor_www_s {
   CURL* curl_handle;
   char error_buffer[CURL_ERROR_SIZE];
   int curl_init_here;
+  int checked_status;
 #endif
 
 #ifdef RAPTOR_WWW_LIBXML
   void *ctxt;
-  xmlGenericErrorFunc old_handler;
   char buffer[RAPTOR_WWW_BUFFER_SIZE];
   int is_end;
+  void *old_xmlGenericErrorContext;
 #endif
 
 #ifdef RAPTOR_WWW_LIBFETCH
@@ -746,9 +900,6 @@ struct  raptor_www_s {
   void *content_type_userdata;
   raptor_www_content_type_handler content_type;
 
-  void *error_data;
-  raptor_message_handler error_handler;
-
   void* uri_filter_user_data;
   raptor_uri_filter_func uri_filter;
 
@@ -758,6 +909,18 @@ struct  raptor_www_s {
   char *http_accept;
 
   FILE* handle;
+
+  raptor_error_handlers error_handlers;
+
+  int connection_timeout;
+
+  /* The URI returned after any redirections */
+  raptor_uri* final_uri;
+
+  void *final_uri_userdata;
+  raptor_www_final_uri_handler final_uri_handler;
+
+  char* cache_control;
 };
 
 
@@ -767,7 +930,6 @@ void raptor_www_libxml_init(raptor_www *www);
 void raptor_www_libxml_free(raptor_www *www);
 int raptor_www_libxml_fetch(raptor_www *www);
 
-void raptor_www_error_varargs(raptor_www *www, const char *message, va_list arguments) RAPTOR_PRINTF_FORMAT(2, 0);
 void raptor_www_error(raptor_www *www, const char *message, ...) RAPTOR_PRINTF_FORMAT(2, 3);
 
 void raptor_www_curl_init(raptor_www *www);
@@ -778,11 +940,11 @@ void raptor_www_libfetch_init(raptor_www *www);
 void raptor_www_libfetch_free(raptor_www *www);
 int raptor_www_libfetch_fetch(raptor_www *www);
 
-  /* raptor_set.c */
-raptor_id_set* raptor_new_id_set(void);
+/* raptor_set.c */
+raptor_id_set* raptor_new_id_set(raptor_world* world);
 void raptor_free_id_set(raptor_id_set* set);
 int raptor_id_set_add(raptor_id_set* set, raptor_uri* base_uri, const unsigned char *item, size_t item_len);
-#ifdef RAPTOR_DEBUG
+#if RAPTOR_DEBUG > 1
 void raptor_id_set_stats_print(raptor_id_set* set, FILE *stream);
 #endif
 
@@ -817,25 +979,11 @@ struct raptor_xml_element_s {
 };
 
 
-/* start of an element */
-typedef void (*raptor_sax2_start_element_handler)(void *user_data, raptor_xml_element *xml_element);
-/* end of an element */
-typedef void (*raptor_sax2_end_element_handler)(void *user_data, raptor_xml_element* xml_element);
-/* characters */
-typedef void (*raptor_sax2_characters_handler)(void *user_data, raptor_xml_element* xml_element, const unsigned char *s, int len);
-/* like <![CDATA[...]> */
-typedef void (*raptor_sax2_cdata_handler)(void *user_data, raptor_xml_element* xml_element, const unsigned char *s, int len);
-/* comment */
-typedef void (*raptor_sax2_comment_handler)(void *user_data, raptor_xml_element* xml_element, const unsigned char *s);
-/* unparsed (NDATA) entity */
-typedef void (*raptor_sax2_unparsed_entity_decl_handler)(void *user_data, const unsigned char* entityName, const unsigned char* base, const unsigned char* systemId, const unsigned char* publicId, const unsigned char* notationName);
-/* external entity reference */
-typedef int (*raptor_sax2_external_entity_ref_handler)(void *user_data, const unsigned char* context, const unsigned char* base, const unsigned char* systemId, const unsigned char* publicId);
-
 struct raptor_sax2_s {
 #ifdef RAPTOR_XML_LIBXML
   int magic;
 #endif
+  raptor_world* world;
   void* user_data;
   
 #ifdef RAPTOR_XML_EXPAT
@@ -857,6 +1005,12 @@ struct raptor_sax2_s {
   int first_read;
 #endif
 
+  void *saved_structured_error_context;
+  xmlStructuredErrorFunc saved_structured_error_handler;
+  
+  void *saved_generic_error_context;
+  xmlGenericErrorFunc saved_generic_error_handler;
+  
 #endif  
 
   /* element depth */
@@ -883,14 +1037,7 @@ struct raptor_sax2_s {
 
   raptor_locator *locator;
 
-  void *error_data;
-  raptor_message_handler error_handler;
-
-  void *fatal_error_data;
-  raptor_message_handler fatal_error_handler;
-
-  void *warning_data;
-  raptor_message_handler warning_handler;
+  raptor_error_handlers* error_handlers;
 
   /* New XML namespace callback */
   raptor_namespace_handler  namespace_handler;
@@ -910,34 +1057,20 @@ struct raptor_sax2_s {
 
   /* base URI for resolving relative URIs or xml:base URIs */
   raptor_uri* base_uri;
+
+  /* sax2 init failed - do not try to do anything with it */
+  int failed;
 };
 
-void raptor_sax2_init(void);
-void raptor_sax2_finish(void);
+int raptor_sax2_init(raptor_world* world);
+void raptor_sax2_finish(raptor_world* world);
 
-raptor_sax2* raptor_new_sax2(void *user_data, void *error_data, raptor_message_handler error_handler, void *fatal_error_data, raptor_message_handler fatal_error_handler, void *warning_data, raptor_message_handler warning_handler);
-void raptor_free_sax2(raptor_sax2 *sax2);
-
-void raptor_sax2_set_start_element_handler(raptor_sax2* sax2, raptor_sax2_start_element_handler handler);
-void raptor_sax2_set_end_element_handler(raptor_sax2* sax2, raptor_sax2_end_element_handler handler);
-void raptor_sax2_set_characters_handler(raptor_sax2* sax2, raptor_sax2_characters_handler handler);
-void raptor_sax2_set_cdata_handler(raptor_sax2* sax2, raptor_sax2_cdata_handler handler);
-void raptor_sax2_set_comment_handler(raptor_sax2* sax2, raptor_sax2_comment_handler handler);
-void raptor_sax2_set_unparsed_entity_decl_handler(raptor_sax2* sax2, raptor_sax2_unparsed_entity_decl_handler handler);
-void raptor_sax2_set_external_entity_ref_handler(raptor_sax2* sax2, raptor_sax2_external_entity_ref_handler handler);
-void raptor_sax2_set_namespace_handler(raptor_sax2* sax2, raptor_namespace_handler handler);
-void raptor_sax2_set_locator(raptor_sax2* sax2, raptor_locator* locator);
-void raptor_sax2_parse_start(raptor_sax2 *sax2, raptor_uri *base_uri);
-int raptor_sax2_parse_chunk(raptor_sax2* sax2, const unsigned char *buffer, size_t len, int is_end);
-void raptor_sax2_parse_handle_errors(raptor_sax2* sax2);
 
 raptor_xml_element* raptor_xml_element_pop(raptor_sax2* sax2);
 void raptor_xml_element_push(raptor_sax2* sax2, raptor_xml_element* element);
 int raptor_sax2_get_depth(raptor_sax2* sax2);
 void raptor_sax2_inc_depth(raptor_sax2* sax2);
 void raptor_sax2_dec_depth(raptor_sax2* sax2);
-const unsigned char* raptor_sax2_inscope_xml_language(raptor_sax2* sax2);
-raptor_uri* raptor_sax2_inscope_base_uri(raptor_sax2* sax2);
 void raptor_sax2_update_document_locator(raptor_sax2* sax2, raptor_locator* locator);
 int raptor_sax2_set_feature(raptor_sax2* sax2, raptor_feature feature, int value);
   
@@ -987,6 +1120,9 @@ struct raptor_uri_detail_s
   size_t path_len;
   size_t query_len;
   size_t fragment_len;
+
+  /* Flags */
+  int is_hierarchical;
 };
 
 
@@ -1008,7 +1144,7 @@ struct raptor_uri_detail_s
 #define RAPTOR_PARSEDATE_FUNCTION parsedate
 #else
 #ifdef HAVE_RAPTOR_PARSE_DATE
-time_t raptor_parse_date(char *p, time_t *now);
+time_t raptor_parse_date(const char *p, time_t *now);
 #define RAPTOR_PARSEDATE_FUNCTION raptor_parse_date
 #else
 #ifdef HAVE_CURL_CURL_H
@@ -1019,16 +1155,17 @@ time_t raptor_parse_date(char *p, time_t *now);
 #endif
 
 /* turtle_common.c */
-int raptor_stringbuffer_append_turtle_string(raptor_stringbuffer* stringbuffer, unsigned char *text, size_t len, int delim, raptor_simple_message_handler error_handler, void *error_data);
+int raptor_stringbuffer_append_turtle_string(raptor_stringbuffer* stringbuffer, const unsigned char *text, size_t len, int delim, raptor_simple_message_handler error_handler, void *error_data);
 
 
 /* raptor_xsd.c */
-raptor_identifier* raptor_new_identifier_from_double(double d);
+raptor_identifier* raptor_new_identifier_from_double(raptor_world* world, double d);
 
 
 /* raptor_abbrev.c */
 
 typedef struct {
+  raptor_world* world;
   int ref_count;         /* count of references to this node */
   int count_as_subject;  /* count of this blank/resource node as subject */
   int count_as_object;   /* count of this blank/resource node as object */
@@ -1062,7 +1199,7 @@ typedef struct {
   raptor_abbrev_node* node;      /* node representing the subject of
                                   * this resource */
   raptor_abbrev_node* node_type; /* the rdf:type of this resource */
-  raptor_sequence *properties;   /* list of properties
+  raptor_avltree *properties;   /* list of properties
                                   * (predicate/object pair) of this
                                   * subject */
   raptor_sequence *list_items;   /* list of container elements if
@@ -1070,18 +1207,19 @@ typedef struct {
 } raptor_abbrev_subject;
 
 
-raptor_abbrev_node* raptor_new_abbrev_node(raptor_identifier_type node_type, const void *node_data, raptor_uri *datatype, const unsigned char *language);
+raptor_abbrev_node* raptor_new_abbrev_node(raptor_world* world, raptor_identifier_type node_type, const void *node_data, raptor_uri *datatype, const unsigned char *language);
 void raptor_free_abbrev_node(raptor_abbrev_node* node);
+int raptor_abbrev_node_cmp(raptor_abbrev_node* node1, raptor_abbrev_node* node2);
 int raptor_abbrev_node_equals(raptor_abbrev_node* node1, raptor_abbrev_node* node2);
 int raptor_abbrev_node_matches(raptor_abbrev_node* node, raptor_identifier_type node_type, const void *node_data, raptor_uri *datatype, const unsigned char *language);
-raptor_abbrev_node* raptor_abbrev_node_lookup(raptor_sequence* nodes, raptor_identifier_type node_type, const void *node_value, raptor_uri *datatype, const unsigned char *language);
+raptor_abbrev_node* raptor_abbrev_node_lookup(raptor_avltree* nodes, raptor_identifier_type node_type, const void *node_value, raptor_uri *datatype, const unsigned char *language, int* created_p);
 
 raptor_abbrev_subject* raptor_new_abbrev_subject(raptor_abbrev_node* node);
 void raptor_free_abbrev_subject(raptor_abbrev_subject* subject);
 int raptor_abbrev_subject_add_property(raptor_abbrev_subject* subject, raptor_abbrev_node* predicate, raptor_abbrev_node* object);
 int raptor_abbrev_subject_add_list_element(raptor_abbrev_subject* subject, int ordinal, raptor_abbrev_node* object);
 raptor_abbrev_subject* raptor_abbrev_subject_find(raptor_sequence *sequence, raptor_identifier_type node_type, const void *node_data, int *idx);
-raptor_abbrev_subject* raptor_abbrev_subject_lookup(raptor_sequence* nodes, raptor_sequence* subjects, raptor_sequence* blanks, raptor_identifier_type node_type, const void *node_data);
+raptor_abbrev_subject* raptor_abbrev_subject_lookup(raptor_avltree* nodes, raptor_sequence* subjects, raptor_sequence* blanks, raptor_identifier_type node_type, const void *node_data, int* created_p);
 
 unsigned char *raptor_unique_id(unsigned char *base);
 
@@ -1096,25 +1234,150 @@ raptor_qname* raptor_new_qname_from_resource(raptor_sequence* namespaces, raptor
 typedef struct raptor_turtle_writer_s raptor_turtle_writer;
 
 /* Turtle Writer Class (raptor_turtle_writer) */
-raptor_turtle_writer* raptor_new_turtle_writer(raptor_uri* base_uri, raptor_namespace_stack *nstack, raptor_uri_handler *uri_handler, void *uri_context, raptor_iostream* iostr, raptor_simple_message_handler error_handler, void *error_data);
+raptor_turtle_writer* raptor_new_turtle_writer(raptor_world* world, raptor_uri* base_uri, int write_base_uri, raptor_namespace_stack *nstack, raptor_iostream* iostr, raptor_simple_message_handler error_handler, void *error_data);
 void raptor_free_turtle_writer(raptor_turtle_writer* turtle_writer);
 void raptor_turtle_writer_raw(raptor_turtle_writer* turtle_writer, const unsigned char *s);
 void raptor_turtle_writer_raw_counted(raptor_turtle_writer* turtle_writer, const unsigned char *s, unsigned int len);
 void raptor_turtle_writer_namespace_prefix(raptor_turtle_writer* turtle_writer, raptor_namespace* ns);
+void raptor_turtle_writer_base(raptor_turtle_writer* turtle_writer, raptor_uri* base_uri);
 void raptor_turtle_writer_increase_indent(raptor_turtle_writer *turtle_writer);
 void raptor_turtle_writer_decrease_indent(raptor_turtle_writer *turtle_writer);
 void raptor_turtle_writer_newline(raptor_turtle_writer *turtle_writer);
 void raptor_turtle_writer_reference(raptor_turtle_writer* turtle_writer, raptor_uri* uri);
-void raptor_turtle_writer_literal(raptor_turtle_writer* turtle_writer, unsigned char *s, unsigned char* lang, raptor_uri* datatype);
+int raptor_turtle_writer_literal(raptor_turtle_writer* turtle_writer, raptor_namespace_stack *nstack, const unsigned char *s, const unsigned char* lang, raptor_uri* datatype);
 void raptor_turtle_writer_qname(raptor_turtle_writer* turtle_writer, raptor_qname* qname);
-void raptor_turtle_writer_quoted(raptor_turtle_writer* turtle_writer, unsigned char *s);
+int raptor_turtle_writer_quoted_counted_string(raptor_turtle_writer* turtle_writer, const unsigned char *s, size_t length);
 void raptor_turtle_writer_comment(raptor_turtle_writer* turtle_writer, const unsigned char *s);
-int raptor_turtle_writer_features_enumerate(const raptor_feature feature, const char **name,  raptor_uri **uri, const char **label);
+int raptor_turtle_writer_features_enumerate(raptor_world* world, const raptor_feature feature, const char **name,  raptor_uri **uri, const char **label);
 int raptor_turtle_writer_set_feature(raptor_turtle_writer *turtle_writer, raptor_feature feature, int value);
 int raptor_turtle_writer_set_feature_string(raptor_turtle_writer *turtle_writer, raptor_feature feature, const unsigned char *value);
 int raptor_turtle_writer_get_feature(raptor_turtle_writer *turtle_writer, raptor_feature feature);
 const unsigned char *raptor_turtle_writer_get_feature_string(raptor_turtle_writer *turtle_writer, raptor_feature feature);
 
+
+/**
+ * raptor_json_writer:
+ *
+ * Raptor JSON Writer class
+ */
+typedef struct raptor_json_writer_s raptor_json_writer;
+
+/* raptor_json_writer.c */
+raptor_json_writer* raptor_new_json_writer(raptor_world* world, raptor_uri* base_uri, raptor_iostream* iostr, raptor_simple_message_handler error_handler, void *error_data);
+void raptor_free_json_writer(raptor_json_writer* json_writer);
+
+int raptor_json_writer_newline(raptor_json_writer* json_writer);
+int raptor_json_writer_key_value(raptor_json_writer* json_writer, const char* key, size_t key_len, const char* value, size_t value_len);
+int raptor_json_writer_start_block(raptor_json_writer* json_writer, char c);
+int raptor_json_writer_end_block(raptor_json_writer* json_writer, char c);
+int raptor_json_writer_literal_object(raptor_json_writer* json_writer, unsigned char* s, unsigned char* lang, raptor_uri* datatype, const char* key, const char* type_key);
+int raptor_json_writer_blank_object(raptor_json_writer* json_writer, const char* blank);
+int raptor_json_writer_uri_object(raptor_json_writer* json_writer, raptor_uri* uri);
+int raptor_json_writer_key_uri_value(raptor_json_writer* json_writer, const char* key, size_t key_len, raptor_uri* uri);
+
+/* raptor_memnstr.c */
+const char* raptor_memstr(const char *haystack, size_t haystack_len, const char *needle);
+
+/* raptor_serialize_rdfxmla.c special functions for embedding rdf/xml */
+int raptor_rdfxmla_serialize_set_write_rdf_RDF(raptor_serializer* serializer, int value);
+int raptor_rdfxmla_serialize_set_xml_writer(raptor_serializer* serializer, raptor_xml_writer* xml_writer, raptor_namespace_stack *nstack);
+int raptor_rdfxmla_serialize_set_single_node(raptor_serializer* serializer, raptor_uri* uri);
+int raptor_rdfxmla_serialize_set_write_typed_nodes(raptor_serializer* serializer, int value);
+
+/* snprintf.c */
+char* raptor_format_float(char *buffer, size_t *currlen, size_t maxlen, double fvalue, unsigned int min, unsigned int max, int flags);
+
+/* constructor / destructor */
+
+#define RAPTOR_AVLTREE_FLAG_REPLACE_DUPLICATES 1
+
+raptor_avltree* raptor_new_avltree(raptor_world* world, raptor_data_compare_function compare_fn, raptor_data_free_function free_fn, unsigned int flags);
+void raptor_free_avltree(raptor_avltree* tree);
+
+/* methods */
+int raptor_avltree_add(raptor_avltree* tree, void* p_user);
+void* raptor_avltree_remove(raptor_avltree* tree, void* p_data);
+int raptor_avltree_delete(raptor_avltree* tree, void* p_user);
+void* raptor_avltree_search(raptor_avltree* tree, const void* p_user);
+int raptor_avltree_visit(raptor_avltree* tree, raptor_avltree_visit_function visit_fn, void* user_data);
+int raptor_avltree_size(raptor_avltree* tree);
+void raptor_avltree_set_print_handler(raptor_avltree* tree, raptor_data_print_function print_fn);
+void raptor_avltree_print(raptor_avltree* tree, FILE* stream);
+
+#ifdef RAPTOR_DEBUG
+int raptor_avltree_dump(raptor_avltree* tree, FILE* stream);
+void raptor_avltree_check(raptor_avltree* tree);
+#endif
+
+int raptor_avltree_cursor_first(raptor_avltree* tree);
+int raptor_avltree_cursor_last(raptor_avltree* tree);
+int raptor_avltree_cursor_prev(raptor_avltree* tree);
+int raptor_avltree_cursor_next(raptor_avltree* tree);
+void* raptor_avltree_cursor_get(raptor_avltree* tree);
+
+raptor_avltree_iterator* raptor_new_avltree_iterator(raptor_avltree* tree, void* range,  raptor_data_free_function range_free_fn, int direction);
+void raptor_free_avltree_iterator(raptor_avltree_iterator* iterator);
+
+int raptor_avltree_iterator_end(raptor_avltree_iterator* iterator);
+int raptor_avltree_iterator_next(raptor_avltree_iterator* iterator);
+void* raptor_avltree_iterator_get(raptor_avltree_iterator* iterator);
+
+/* raptor_world structure */
+struct raptor_world_s {
+  /* world has been initialized with raptor_world_open() */
+  int opened;
+
+  /* raptor_init(), raptor_finish() balance */
+  int static_usage;
+
+  /* sequence of parser factories */
+  raptor_sequence *parsers;
+
+  /* sequence of serializer factories */
+  raptor_sequence *serializers;
+
+  /* current uri handler and its data */
+  const raptor_uri_handler *uri_handler;
+  void *uri_handler_context;
+
+  /* raptor_rss_common initialisation counter */
+  int rss_common_initialised;
+
+  /* raptor_rss_{namespaces,types,fields}_info const data initialized to raptor_uri,raptor_qname objects */
+  raptor_uri **rss_namespaces_info_uris;
+  raptor_uri **rss_types_info_uris;
+  raptor_qname **rss_types_info_qnames;
+  raptor_uri **rss_fields_info_uris;
+  raptor_qname **rss_fields_info_qnames;
+
+  /* raptor_www v2 flags */
+  int www_skip_www_init_finish;
+  int www_initialized;
+
+  /* raptor_sax2 init counter to work around issues in xml parser init/cleanup */
+  int sax2_initialized;
+
+  /* This is used to store a #xsltSecurityPrefsPtr typed object
+   * pointer when libxslt is compiled in.
+   */
+  void* xslt_security_preferences;
+  /* If non-0 - raptors own the above object and should free it with
+   * xsltFreeSecurityPrefs() on exit
+   */
+  int free_xslt_security_preferences;
+
+  /* Flags for libxml set by raptor_world_set_libxml_flags() and
+   * raptor_set_libxml_flags() - see #raptor_libxml_flags for
+   * meanings 
+   */
+  int libxml_flags;
+};
+
+/* raptor_world legacy accessor */
+#ifndef RAPTOR_DISABLE_V1
+raptor_world* raptor_world_instance(void);
+#endif
+
 /* end of RAPTOR_INTERNAL */
 #endif
 
