@@ -2,9 +2,7 @@
  *
  * raptor_namespace.c - Raptor XML namespace classes
  *
- * $Id: raptor_namespace.c,v 1.62 2006/01/03 03:11:49 cmdjb Exp $
- *
- * Copyright (C) 2002-2006, David Beckett http://purl.org/net/dajobe/
+ * Copyright (C) 2002-2009, David Beckett http://www.dajobe.org/
  * Copyright (C) 2002-2005, University of Bristol, UK http://www.bristol.ac.uk/
  * 
  * This package is Free Software and part of Redland http://librdf.org/
@@ -99,11 +97,12 @@ const unsigned char * const raptor_xmlschema_datatypes_namespace_uri=(const unsi
 const unsigned char * const raptor_owl_namespace_uri=(const unsigned char *)"http://www.w3.org/2002/07/owl#";
 
 
+#ifndef RAPTOR_DISABLE_V1
 /**
  * raptor_namespaces_init:
  * @nstack: #raptor_namespace_stack to initialise
- * @uri_handler: URI handler function
- * @uri_context: context for URI handler
+ * @uri_handler: URI handler function (ignored)
+ * @uri_context: context for URI handler (ignored)
  * @error_handler: error handler function
  * @error_data: context for error handler
  * @defaults: namespaces to initialise.
@@ -112,76 +111,196 @@ const unsigned char * const raptor_owl_namespace_uri=(const unsigned char *)"htt
  *
  * @defaults can be 0 for none, 1 for just XML, 2 for RDF, RDFS, OWL
  * and XSD (RDQL uses this) or 3+ undefined.
+ *
+ * @uri_handler and @uri_context parameters are ignored but are retained
+ * in the API for backwards compatibility. Internally the same uri handler
+ * as returned by raptor_uri_get_handler() will be used.
+ *
+ * raptor_init() MUST have been called before calling this function.
+ * Use raptor_namespaces_init_v2() if using raptor_world APIs.
+ *
+ * Return value: non-0 on error
  */
-void
+int
 raptor_namespaces_init(raptor_namespace_stack *nstack,
-                       raptor_uri_handler *uri_handler,
+                       const raptor_uri_handler *uri_handler,
                        void *uri_context,
                        raptor_simple_message_handler error_handler,
                        void *error_data,
                        int defaults)
 {
-  nstack->top=NULL;
-  nstack->uri_handler=uri_handler;
-  nstack->uri_context=uri_context;
+  return raptor_namespaces_init_v2(raptor_world_instance(),
+                                   nstack,
+                                   error_handler,
+                                   error_data,
+                                   defaults);
+}
+#endif
+
+
+/* hash function to hash namespace prefix strings (usually short strings)
+ *
+ * Uses DJ Bernstein original hash function - good on short text keys.
+ */
+static unsigned int
+raptor_hash_ns_string(const unsigned char *str, int length)
+{
+  unsigned int hash = 5381;
+  int c;
+  
+  for(; length && (c = *str++); length--) {
+    hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
+  }
+
+  return hash;
+}
+
+
+#define RAPTOR_NAMESPACES_HASHTABLE_SIZE 1024
+/**
+ * raptor_namespaces_init_v2:
+ * @world: raptor_world object
+ * @nstack: #raptor_namespace_stack to initialise
+ * @error_handler: error handler function
+ * @error_data: context for error handler
+ * @defaults: namespaces to initialise.
+ *
+ * Initialise a namespaces stack some optional common namespaces.
+ *
+ * @defaults can be 0 for none, 1 for just XML, 2 for RDF, RDFS, OWL
+ * and XSD (RDQL uses this) or 3+ undefined.
+ *
+ * Return value: non-0 on error
+ */
+int
+raptor_namespaces_init_v2(raptor_world* world,
+                          raptor_namespace_stack *nstack,
+                          raptor_simple_message_handler error_handler,
+                          void *error_data,
+                          int defaults)
+{
+  int failures=0;
+
+  nstack->world=world;
 
   nstack->error_handler=error_handler;
   nstack->error_data=error_data;
 
-  nstack->rdf_ms_uri    = uri_handler->new_uri(uri_context, (const unsigned char*)raptor_rdf_namespace_uri);
-  nstack->rdf_schema_uri= uri_handler->new_uri(uri_context, (const unsigned char*)raptor_rdf_schema_namespace_uri);
+  nstack->size = 0;
+  
+  nstack->table_size = RAPTOR_NAMESPACES_HASHTABLE_SIZE;
+  nstack->table = RAPTOR_CALLOC(raptor_namespaces,
+                                RAPTOR_NAMESPACES_HASHTABLE_SIZE,
+                                sizeof(raptor_namespace*));
+  if(!nstack->table)
+    return -1;
 
-  if(defaults) {
+  nstack->def_namespace = NULL;
+
+  nstack->rdf_ms_uri = raptor_new_uri_v2(nstack->world,
+                                         (const unsigned char*)raptor_rdf_namespace_uri);
+  failures += !nstack->rdf_ms_uri;
+   
+  nstack->rdf_schema_uri = raptor_new_uri_v2(nstack->world,
+                                             (const unsigned char*)raptor_rdf_schema_namespace_uri);
+  failures += !nstack->rdf_schema_uri;
+
+  /* raptor_new_namespace_from_uri() that eventually gets called by
+   * raptor_new_namespace() in raptor_namespaces_start_namespace_full()
+   * needs rdf_ms_uri and rdf_schema_uri
+   * - do not call if we had failures initializing those uris */
+  if(defaults && !failures) {
     /* defined at level -1 since always 'present' when inside the XML world */
-    raptor_namespaces_start_namespace_full(nstack, (const unsigned char*)"xml",
-                                           raptor_xml_namespace_uri, -1);
+    failures += raptor_namespaces_start_namespace_full(nstack,
+                                                       (const unsigned char*)"xml",
+                                                       raptor_xml_namespace_uri, -1);
     if(defaults >= 2) {
-      raptor_namespaces_start_namespace_full(nstack,
-                                             (const unsigned char*)"rdf",
-                                             raptor_rdf_namespace_uri, 0);
-      raptor_namespaces_start_namespace_full(nstack,
-                                             (const unsigned char*)"rdfs",
-                                             raptor_rdf_schema_namespace_uri, 0);
-      raptor_namespaces_start_namespace_full(nstack,
-                                             (const unsigned char*)"xsd",
-                                             raptor_xmlschema_datatypes_namespace_uri, 0);
-      raptor_namespaces_start_namespace_full(nstack,
-                                             (const unsigned char*)"owl",
-                                             raptor_owl_namespace_uri, 0);
+      failures += raptor_namespaces_start_namespace_full(nstack,
+                                                         (const unsigned char*)"rdf",
+                                                         raptor_rdf_namespace_uri, 0);
+      failures += raptor_namespaces_start_namespace_full(nstack,
+                                                         (const unsigned char*)"rdfs",
+                                                         raptor_rdf_schema_namespace_uri, 0);
+      failures += raptor_namespaces_start_namespace_full(nstack,
+                                                         (const unsigned char*)"xsd",
+                                                         raptor_xmlschema_datatypes_namespace_uri, 0);
+      failures += raptor_namespaces_start_namespace_full(nstack,
+                                                         (const unsigned char*)"owl",
+                                                         raptor_owl_namespace_uri, 0);
     }
   }
+  return failures;
 }
 
 
+#ifndef RAPTOR_DISABLE_V1
 /**
  * raptor_new_namespaces:
- * @uri_handler: URI handler function
- * @uri_context: URI handler context data
+ * @uri_handler: URI handler function (ignored)
+ * @uri_context: URI handler context data (ignored)
  * @error_handler: error handler function
- * @error_data: erorr handler data
+ * @error_data: error handler data
  * @defaults: namespaces to initialise
  * 
  * Constructor - create a new #raptor_namespace_stack.
  *
- * See raptor_namespaces_init() fo the values of @defaults.
+ * See raptor_namespaces_init() for the values of @defaults.
+ *
+ * @uri_handler and @uri_context parameters are ignored but are retained
+ * in the API for backwards compatibility. Internally the same uri handler
+ * as returned by raptor_uri_get_handler() will be used.
+ *
+ * raptor_init() MUST have been called before calling this function.
+ * Use raptor_new_namespaces_v2() if using raptor_world APIs.
  * 
  * Return value: a new namespace stack or NULL on failure
  **/
 raptor_namespace_stack *
-raptor_new_namespaces(raptor_uri_handler *uri_handler,
+raptor_new_namespaces(const raptor_uri_handler *uri_handler,
                       void *uri_context,
                       raptor_simple_message_handler error_handler,
                       void *error_data,
                       int defaults) 
 {
-  raptor_namespace_stack *nstack=(raptor_namespace_stack *)RAPTOR_MALLOC(raptor_namespace_stack, sizeof(raptor_namespace_stack));
+  return raptor_new_namespaces_v2(raptor_world_instance(),
+                                  error_handler,
+                                  error_data,
+                                  defaults);
+}
+#endif
+
+
+/**
+ * raptor_new_namespaces_v2:
+ * @world: raptor_world object
+ * @error_handler: error handler function
+ * @error_data: error handler data
+ * @defaults: namespaces to initialise
+ * 
+ * Constructor - create a new #raptor_namespace_stack.
+ *
+ * See raptor_namespaces_init() for the values of @defaults.
+ * 
+ * Return value: a new namespace stack or NULL on failure
+ **/
+raptor_namespace_stack *
+raptor_new_namespaces_v2(raptor_world* world,
+                         raptor_simple_message_handler error_handler,
+                         void *error_data,
+                         int defaults) 
+{
+  raptor_namespace_stack *nstack=(raptor_namespace_stack *)RAPTOR_CALLOC(raptor_namespace_stack, 1, sizeof(raptor_namespace_stack));
   if(!nstack)
     return NULL;
                       
-  raptor_namespaces_init(nstack, 
-                         uri_handler, uri_context,
-                         error_handler, error_data,
-                         defaults);
+  if(raptor_namespaces_init_v2(world,
+                               nstack, 
+                               error_handler, error_data,
+                               defaults)) {
+    raptor_free_namespaces(nstack);
+    nstack=NULL;
+  }
+
   return nstack;
 }
  
@@ -197,9 +316,18 @@ void
 raptor_namespaces_start_namespace(raptor_namespace_stack *nstack, 
                                   raptor_namespace *nspace)
 {
-  if(nstack->top)
-    nspace->next=nstack->top;
-  nstack->top=nspace;
+  unsigned int hash = raptor_hash_ns_string(nspace->prefix,
+                                            nspace->prefix_length);
+  const int bucket = hash % nstack->table_size;
+
+  nstack->size++;
+  
+  if(nstack->table[bucket])
+    nspace->next = nstack->table[bucket];
+  nstack->table[bucket] = nspace;
+
+  if(!nstack->def_namespace)
+    nstack->def_namespace = nspace;
 
 #ifndef STANDALONE
 #ifdef RAPTOR_DEBUG_VERBOSE
@@ -232,7 +360,7 @@ raptor_namespaces_start_namespace_full(raptor_namespace_stack *nstack,
 {
   raptor_namespace *ns;
 
-  ns=raptor_new_namespace(nstack, prefix, ns_uri_string, depth);
+  ns = raptor_new_namespace(nstack, prefix, ns_uri_string, depth);
   if(!ns)
     return 1;
   
@@ -250,22 +378,40 @@ raptor_namespaces_start_namespace_full(raptor_namespace_stack *nstack,
 void
 raptor_namespaces_clear(raptor_namespace_stack *nstack)
 {
-  raptor_namespace *ns=nstack->top;
-  while(ns) {
-    raptor_namespace* next_ns=ns->next;
+  if(nstack->table) {
+    int bucket;
 
-    raptor_free_namespace(ns);
-    ns=next_ns;
-  }
-  nstack->top=NULL;
+    for(bucket = 0; bucket < nstack->table_size; bucket++) {
+      raptor_namespace *ns = nstack->table[bucket];
+      while(ns) {
+        raptor_namespace* next_ns = ns->next;
+        
+        raptor_free_namespace(ns);
+        nstack->size--;
+        ns = next_ns;
+      }
+      nstack->table[bucket] = NULL;
+    }
 
-  if(nstack->uri_handler) {
-    nstack->uri_handler->free_uri(nstack->uri_context, nstack->rdf_ms_uri);
-    nstack->uri_handler->free_uri(nstack->uri_context, nstack->rdf_schema_uri);
+    RAPTOR_FREE(raptor_namespaces, nstack->table);
+    nstack->table = NULL;
+    nstack->table_size = 0;
   }
 
-  nstack->uri_handler=(raptor_uri_handler*)NULL;
-  nstack->uri_context=NULL;
+  if(nstack->world) {
+    if(nstack->rdf_ms_uri) {
+      raptor_free_uri_v2(nstack->world, nstack->rdf_ms_uri);
+      nstack->rdf_ms_uri = NULL;
+    }
+    if(nstack->rdf_schema_uri) {
+      raptor_free_uri_v2(nstack->world, nstack->rdf_schema_uri);
+      nstack->rdf_schema_uri = NULL;
+    }
+  }
+
+  nstack->size = 0;
+
+  nstack->world = NULL;
 }
 
 
@@ -278,7 +424,11 @@ raptor_namespaces_clear(raptor_namespace_stack *nstack)
 void
 raptor_free_namespaces(raptor_namespace_stack *nstack)
 {
+  if(!nstack)
+    return;
+  
   raptor_namespaces_clear(nstack);
+
   RAPTOR_FREE(raptor_namespace_stack, nstack);
 }
 
@@ -293,21 +443,25 @@ raptor_free_namespaces(raptor_namespace_stack *nstack)
 void 
 raptor_namespaces_end_for_depth(raptor_namespace_stack *nstack, int depth)
 {
-  while(nstack->top &&
-        nstack->top->depth == depth) {
-    raptor_namespace* ns=nstack->top;
-    raptor_namespace* next=ns->next;
+  int bucket;
+  for(bucket = 0; bucket < nstack->table_size; bucket++) {
+    while(nstack->table[bucket] &&
+          nstack->table[bucket]->depth == depth) {
+      raptor_namespace* ns = nstack->table[bucket];
+      raptor_namespace* next_ns = ns->next;
 
 #ifndef STANDALONE
 #ifdef RAPTOR_DEBUG_VERBOSE
-    RAPTOR_DEBUG3("namespace prefix %s depth %d\n", ns->prefix ? (char*)ns->prefix : "(default)", depth);
+      RAPTOR_DEBUG3("namespace prefix %s depth %d\n",
+                    ns->prefix ? (char*)ns->prefix : "(default)", depth);
 #endif
 #endif
-    raptor_free_namespace(ns);
+      raptor_free_namespace(ns);
+      nstack->size--;
 
-    nstack->top=next;
+      nstack->table[bucket] = next_ns;
+    }
   }
-
 }
 
 
@@ -322,9 +476,11 @@ raptor_namespaces_end_for_depth(raptor_namespace_stack *nstack, int depth)
 raptor_namespace*
 raptor_namespaces_get_default_namespace(raptor_namespace_stack *nstack)
 {
+  unsigned int hash = raptor_hash_ns_string((const unsigned char *)"", 0);
+  const int bucket = hash % nstack->table_size;
   raptor_namespace* ns;
-  
-  for(ns=nstack->top; ns && ns->prefix; ns=ns->next)
+
+  for(ns = nstack->table[bucket]; ns && ns->prefix; ns = ns->next)
     ;
   return ns;
 }
@@ -349,8 +505,10 @@ raptor_namespaces_find_namespace(raptor_namespace_stack *nstack,
                                  const unsigned char *prefix, int prefix_length)
 {
   raptor_namespace* ns;
-  
-  for(ns=nstack->top; ns ; ns=ns->next) {
+  unsigned int hash = raptor_hash_ns_string(prefix, prefix_length);
+  const int bucket = hash % (nstack->table_size);
+
+  for(ns = nstack->table[bucket]; ns ; ns = ns->next) {
     if(!prefix && !ns->prefix)
       break;
     
@@ -358,6 +516,7 @@ raptor_namespaces_find_namespace(raptor_namespace_stack *nstack,
        !strncmp((char*)prefix, (char*)ns->prefix, prefix_length))
       break;
   }
+
   return ns;
 }
 
@@ -375,15 +534,18 @@ raptor_namespace*
 raptor_namespaces_find_namespace_by_uri(raptor_namespace_stack *nstack, 
                                         raptor_uri *ns_uri)
 {
-  raptor_namespace* ns;
+  int bucket;
 
   if(!ns_uri)
     return NULL;
   
-  for(ns=nstack->top; ns ; ns=ns->next)
-    if(nstack->uri_handler->uri_equals(nstack->uri_context, ns->uri, ns_uri))
-      return ns;
-
+  for (bucket = 0; bucket < nstack->table_size; bucket++) {
+    raptor_namespace* ns;
+    for(ns = nstack->table[bucket]; ns ; ns = ns->next)
+      if(raptor_uri_equals_v2(nstack->world, ns->uri, ns_uri))
+        return ns;
+   }
+  
   return NULL;
 }
 
@@ -402,10 +564,13 @@ raptor_namespaces_namespace_in_scope(raptor_namespace_stack *nstack,
                                      const raptor_namespace *nspace)
 {
   raptor_namespace* ns;
+  int bucket;
   
-  for(ns=nstack->top; ns ; ns=ns->next)
-    if(nstack->uri_handler->uri_equals(nstack->uri_context, ns->uri, nspace->uri))
-      return 1;
+  for(bucket = 0; bucket < nstack->table_size; bucket++) {
+    for(ns = nstack->table[bucket]; ns ; ns = ns->next)
+      if(raptor_uri_equals_v2(nstack->world, ns->uri, nspace->uri))
+        return 1;
+  }
   return 0;
 }
 
@@ -433,7 +598,10 @@ raptor_new_namespace_from_uri(raptor_namespace_stack *nstack,
 
 #ifndef STANDALONE
 #if RAPTOR_DEBUG >1
-  RAPTOR_DEBUG4("namespace prefix %s uri %s depth %d\n", prefix ? (char*)prefix : "(default)", ns_uri ? (char*)raptor_uri_as_string(ns_uri) : "(none)", depth);
+  RAPTOR_DEBUG4("namespace prefix %s uri %s depth %d\n",
+                prefix ? (char*)prefix : "(default)",
+                ns_uri ? (char*)raptor_uri_as_string_v2(nstack->world, ns_uri) : "(none)",
+                depth);
 #endif
 #endif
 
@@ -458,7 +626,7 @@ raptor_new_namespace_from_uri(raptor_namespace_stack *nstack,
 
   p=(unsigned char*)ns+sizeof(raptor_namespace);
   if(ns_uri) {
-    ns->uri=(*nstack->uri_handler->uri_copy)(nstack->uri_context, ns_uri);
+    ns->uri = raptor_uri_copy_v2(nstack->world, ns_uri);
     if(!ns->uri) {
       RAPTOR_FREE(raptor_namespace, ns);
       return NULL;
@@ -475,9 +643,9 @@ raptor_new_namespace_from_uri(raptor_namespace_stack *nstack,
 
   /* set convienience flags when there is a defined namespace URI */
   if(ns->uri) {
-    if(nstack->uri_handler->uri_equals(nstack->uri_context, ns->uri, nstack->rdf_ms_uri))
-      ns->is_rdf_ms=1;
-    else if(nstack->uri_handler->uri_equals(nstack->uri_context, ns->uri, nstack->rdf_schema_uri))
+    if(raptor_uri_equals_v2(nstack->world, ns->uri, nstack->rdf_ms_uri))
+      ns->is_rdf_ms = 1;
+    else if(raptor_uri_equals_v2(nstack->world, ns->uri, nstack->rdf_schema_uri))
       ns->is_rdf_schema=1;
   }
 
@@ -510,11 +678,14 @@ raptor_new_namespace(raptor_namespace_stack *nstack,
   if(ns_uri_string && !*ns_uri_string)
     ns_uri_string=NULL;
 
-  if(ns_uri_string)
-    ns_uri=raptor_new_uri(ns_uri_string);
+  if(ns_uri_string) {
+    ns_uri=raptor_new_uri_v2(nstack->world, ns_uri_string);
+    if(!ns_uri)
+      return NULL;
+  }
   ns=raptor_new_namespace_from_uri(nstack, prefix, ns_uri, depth);
   if(ns_uri)
-    raptor_free_uri(ns_uri);
+    raptor_free_uri_v2(nstack->world, ns_uri);
 
   return ns;
 }
@@ -555,8 +726,10 @@ raptor_namespace_copy(raptor_namespace_stack *nstack,
 void 
 raptor_free_namespace(raptor_namespace *ns)
 {
+  RAPTOR_ASSERT_OBJECT_POINTER_RETURN(ns, raptor_namespace);
+
   if(ns->uri)
-    ns->nstack->uri_handler->free_uri(ns->nstack->uri_context, ns->uri);
+    raptor_free_uri_v2(ns->nstack->world, ns->uri);
 
   RAPTOR_FREE(raptor_namespace, ns);
 }
@@ -592,6 +765,24 @@ raptor_namespace_get_prefix(const raptor_namespace *ns)
 }
 
 
+/**
+ * raptor_namespace_get_counted_prefix:
+ * @ns: namespace object
+ * @length_p: pointer to store length or NULL
+ * 
+ * Get the namespace prefix and length.
+ * 
+ * Return value: prefix string or NULL
+ **/
+const unsigned char*
+raptor_namespace_get_counted_prefix(const raptor_namespace *ns, size_t *length_p)
+{
+  if(length_p)
+    *length_p=ns->prefix_length;
+  return (const unsigned char*)ns->prefix;
+}
+
+
 /**
  * raptor_namespaces_format:
  * @ns: namespace object
@@ -622,7 +813,7 @@ raptor_namespaces_format(const raptor_namespace *ns, size_t *length_p)
   unsigned char *p;
 
   if(ns->uri) {
-    uri_string=raptor_uri_as_counted_string(ns->uri, &uri_length);
+    uri_string=raptor_uri_as_counted_string_v2(ns->nstack->world, ns->uri, &uri_length);
     xml_uri_length=raptor_xml_escape_string(uri_string, uri_length,
                                             NULL, 0, quote, NULL, NULL);
   }
@@ -658,6 +849,7 @@ raptor_namespaces_format(const raptor_namespace *ns, size_t *length_p)
     p+= xml_uri_length;
   }
   *p++ = quote;
+  *p++ = '\0';
 
   return buffer;
 }
@@ -682,7 +874,7 @@ raptor_iostream_write_namespace(raptor_iostream* iostr, raptor_namespace *ns)
     return 1;
   
   if(ns->uri)
-    uri_string=raptor_uri_as_counted_string(ns->uri, &uri_length);
+    uri_string=raptor_uri_as_counted_string_v2(ns->nstack->world, ns->uri, &uri_length);
   
   raptor_iostream_write_counted_string(iostr, "xmlns", 5);
   if(ns->prefix) {
@@ -795,11 +987,77 @@ raptor_new_namespace_parts_from_string(const unsigned char *string,
 }
 
 
+/**
+ * raptor_namespaces_qname_from_uri:
+ * @nstack: namespace stack
+ * @uri: URI to use to make qname
+ * @xml_version: XML Version
+ * 
+ * Make an appropriate XML Qname from the namespaces on a namespace stack
+ * 
+ * Makes a qname from the in-scope namespaces in a stack if the URI matches
+ * the prefix and the rest is a legal XML name.
+ *
+ * Return value: #raptor_qname for the URI or NULL on failure
+ **/
+raptor_qname*
+raptor_namespaces_qname_from_uri(raptor_namespace_stack *nstack, 
+                                 raptor_uri *uri, int xml_version)
+{
+  unsigned char *uri_string;
+  size_t uri_len;
+  raptor_namespace* ns = NULL;
+  unsigned char *ns_uri_string;
+  size_t ns_uri_len;
+  unsigned char *name = NULL;
+  int bucket;
+
+  if(!uri)
+    return NULL;
+  
+  uri_string = raptor_uri_as_counted_string_v2(nstack->world, uri, &uri_len);
+
+  for(bucket = 0; bucket < nstack->table_size; bucket++) {
+    for(ns = nstack->table[bucket]; ns ; ns = ns->next) {
+      if(!ns->uri)
+        continue;
+      
+      ns_uri_string = raptor_uri_as_counted_string_v2(nstack->world, ns->uri,
+                                                      &ns_uri_len);
+      if(ns_uri_len >= uri_len)
+        continue;
+      if(strncmp((const char*)uri_string, (const char*)ns_uri_string,
+                 ns_uri_len))
+        continue;
+      
+      /* uri_string is a prefix of ns_uri_string */
+      name = uri_string + ns_uri_len;
+      if(!raptor_xml_name_check(name, uri_len-ns_uri_len, xml_version))
+        name = NULL;
+      
+      /* If name is set, we've found a prefix with a legal XML name value */
+      if(name)
+        break;
+    }
+    if(name)
+      break;
+  }
+  
+  if(!ns)
+    return NULL;
+
+  return raptor_new_qname_from_namespace_local_name_v2(nstack->world, ns,
+                                                       name,  NULL);
+}
+
+
 #ifdef RAPTOR_DEBUG
 void
 raptor_namespace_print(FILE *stream, raptor_namespace* ns) 
 {
-  const unsigned char *uri_string=raptor_uri_as_string(ns->uri);
+  const unsigned char *uri_string;
+
+  uri_string = raptor_uri_as_string_v2(ns->nstack->world, ns->uri);
   if(ns->prefix)
     fprintf(stream, "%s:%s", ns->prefix, uri_string);
   else
@@ -808,6 +1066,49 @@ raptor_namespace_print(FILE *stream, raptor_namespace* ns)
 #endif
 
 
+raptor_namespace**
+raptor_namespace_stack_to_array(raptor_namespace_stack *nstack,
+                                size_t *size_p)
+{
+  raptor_namespace** ns_list;
+  size_t size = 0;
+  int bucket;
+  
+  ns_list = RAPTOR_CALLOC(namespace_stack, nstack->size,
+                          sizeof(raptor_namespace*));
+  if(!ns_list)
+    return NULL;
+  
+  for(bucket = 0; bucket < nstack->table_size; bucket++) {
+    raptor_namespace* ns;
+
+    for(ns = nstack->table[bucket]; ns; ns = ns->next) {
+      int skip = 0;
+      unsigned int i;
+      if(ns->depth < 1)
+        continue;
+      
+      for(i = 0; i < size; i++) {
+        raptor_namespace* ns2 = ns_list[i];
+        if((!ns->prefix && !ns2->prefix) ||
+           (ns->prefix && ns2->prefix && 
+            !strcmp((const char*)ns->prefix, (const char*)ns2->prefix))) {
+          /* this prefix was seen (overridden) earlier so skip */
+          skip = 1;
+          break;
+        }
+      }
+      if(!skip)
+        ns_list[size++] = ns;
+    }
+  }
+  
+  if(size_p)
+    *size_p = size;
+
+  return ns_list;
+}
+
 
 #ifdef STANDALONE
 
@@ -819,18 +1120,16 @@ int main(int argc, char *argv[]);
 int
 main(int argc, char *argv[]) 
 {
+  raptor_world *world;
   const char *program=raptor_basename(argv[0]);
   raptor_namespace_stack namespaces;
   raptor_namespace* ns;
-  raptor_uri_handler *handler;
-  void *context;
 
-  raptor_uri_init();
+  world = raptor_new_world();
+  if(!world || raptor_world_open(world))
+    exit(1);
 
-  /* Use whatever the raptor_uri class has */
-  raptor_uri_get_handler(&handler, &context);
-
-  raptor_namespaces_init(&namespaces, handler, context, NULL, NULL, 1);
+  raptor_namespaces_init_v2(world, &namespaces, NULL, NULL, 1);
   
   raptor_namespaces_start_namespace_full(&namespaces,
                                          (const unsigned char*)"ex1",
@@ -875,6 +1174,8 @@ main(int argc, char *argv[])
 
   raptor_namespaces_clear(&namespaces);
 
+  raptor_free_world(world);
+
   /* keep gcc -Wall happy */
   return(0);
 }
