@@ -2,7 +2,7 @@
  *
  * raptor_www.c - Raptor WWW retrieval core
  *
- * Copyright (C) 2003-2006, David Beckett http://purl.org/net/dajobe/
+ * Copyright (C) 2003-2008, David Beckett http://www.dajobe.org/
  * Copyright (C) 2003-2005, University of Bristol, UK http://www.bristol.ac.uk/
  * 
  * This package is Free Software and part of Redland http://librdf.org/
@@ -46,37 +46,86 @@
 #include "raptor_internal.h"
 
 
-
+static int raptor_www_init_common(int skip_www_init_finish, int *www_initialized);
+static void raptor_www_finish_common(int skip_www_init_finish);
 static int raptor_www_file_fetch(raptor_www* www);
 
 
+
+#ifndef RAPTOR_DISABLE_V1
 /* should raptor_www do initializing and cleanup of the WWW library */
-static int raptor_www_do_www_init_finish=1;
+static int raptor_www_skip_www_init_finish=0;
+static int raptor_www_initialized=0;
+#endif
 
 
+#ifndef RAPTOR_DISABLE_V1
 /**
  * raptor_www_init:
  * 
  * Initialise the WWW class.
  *
  * Must be called before creating any #raptor_www object.
+ *
+ * See also: raptor_www_init_v2()
  **/
 void
 raptor_www_init(void)
 {
-  static int initialized=0;
-  if(initialized)
-    return;
-
-  if(raptor_www_do_www_init_finish) {
-#ifdef RAPTOR_WWW_LIBCURL
-    curl_global_init(CURL_GLOBAL_ALL);
+  raptor_www_init_common(raptor_www_skip_www_init_finish, &raptor_www_initialized);
+}
 #endif
-  }
-  initialized=1;
+
+
+/**
+ * raptor_www_init_v2:
+ * @world: raptor_world object
+ * 
+ * Initialise the WWW class.
+ *
+ * Must be called before creating any #raptor_www object.
+ *
+ * See also: raptor_www_init()
+ *
+ * Return value: non-0 on failure
+ **/
+int
+raptor_www_init_v2(raptor_world* world)
+{
+#ifndef RAPTOR_DISABLE_V1
+  /* support legacy v1 raptor_www_no_www_library_init_finish() */
+  if(raptor_www_skip_www_init_finish)
+    world->www_skip_www_init_finish = raptor_www_skip_www_init_finish;
+
+  /* skip init if already inited with legacy init() */
+  if(raptor_www_initialized)
+    return 0;
+#endif
+
+  return raptor_www_init_common(world->www_skip_www_init_finish, &world->www_initialized);
 }
 
 
+static int
+raptor_www_init_common(int skip_www_init_finish, int *www_initialized)
+{
+  int rc = 0;
+
+  if(*www_initialized)
+    return 0;
+
+  if(!skip_www_init_finish) {
+#ifdef RAPTOR_WWW_LIBCURL
+    rc = curl_global_init(CURL_GLOBAL_ALL);
+#endif
+  }
+
+  *www_initialized = 1;
+  return rc;
+}
+
+
+#ifndef RAPTOR_DISABLE_V1
 /**
  * raptor_www_no_www_library_init_finish:
  *
@@ -95,25 +144,83 @@ raptor_www_init(void)
  * 
  * This function must be called before raptor_init.
  *
+ * See also: raptor_www_no_www_library_init_finish_v2()
+ *
  **/
 void
 raptor_www_no_www_library_init_finish(void)
 {
-  raptor_www_do_www_init_finish=0;
+  raptor_www_skip_www_init_finish = 1;
+}
+#endif
+
+
+/**
+ * raptor_www_no_www_library_init_finish_v2:
+ * @world: raptor_world object
+ *
+ * Do not initialise or finish the lower level WWW library.
+ *
+ * If this is called then the raptor_www library will neither
+ * initialise or terminate the lower level WWW library.  Usually in
+ * raptor_world_open() either curl_global_init (for libcurl)
+ * are called and in raptor_finish curl_global_cleanup is called.
+ *
+ * This allows the application finer control over these libraries such
+ * as setting other global options or potentially calling and terminating
+ * raptor several times.  It does mean that applications which use
+ * this call must do their own extra work in order to allocate and free
+ * all resources to the system.
+ * 
+ * This function must be called before raptor_world_open().
+ *
+ **/
+void
+raptor_www_no_www_library_init_finish_v2(raptor_world* world)
+{
+  world->www_skip_www_init_finish = 1;
 }
 
 
+#ifndef RAPTOR_DISABLE_V1
 /**
  * raptor_www_finish:
  * 
  * Terminate the WWW class.
  *
  * Must be called to clean any resources used by the WWW implementation.
+ *
+ * See also: raptor_www_finish_v2()
  **/
 void
 raptor_www_finish(void)
 {
-  if(raptor_www_do_www_init_finish) {
+  raptor_www_finish_common(raptor_www_skip_www_init_finish);
+}
+#endif
+
+
+/**
+ * raptor_www_finish_v2:
+ * @world: raptor_world object
+ * 
+ * Terminate the WWW class.
+ *
+ * Must be called to clean any resources used by the WWW implementation.
+ *
+ * See also: raptor_www_finish()
+ **/
+void
+raptor_www_finish_v2(raptor_world* world)
+{
+  raptor_www_finish_common(world->www_skip_www_init_finish);
+}
+
+
+static void
+raptor_www_finish_common(int skip_www_init_finish)
+{
+  if(!skip_www_init_finish) {
 #ifdef RAPTOR_WWW_LIBCURL
     curl_global_cleanup();
 #endif
@@ -121,7 +228,7 @@ raptor_www_finish(void)
 }
 
 
-
+#ifndef RAPTOR_DISABLE_V1
 /**
  * raptor_www_new_with_connection:
  * @connection: external WWW connection object.
@@ -132,16 +239,42 @@ raptor_www_finish(void)
  * when raptor is compiled with libcurl suppport. Otherwise the
  * @connection is ignored.  This allows such things as setting
  * up special flags on the curl handle before passing into the constructor.
+ *
+ * raptor_init() MUST have been called before calling this function.
+ * Use raptor_www_new_with_connection_v2() if using raptor_world APIs.
+ * 
+ * Return value: a new #raptor_www object or NULL on failure.
+ **/
+raptor_www* 
+raptor_www_new_with_connection(void *connection)
+{
+  return raptor_www_new_with_connection_v2(raptor_world_instance(), connection);
+}
+#endif
+
+
+/**
+ * raptor_www_new_with_connection_v2:
+ * @world: raptor_world object
+ * @connection: external WWW connection object.
+ * 
+ * Constructor - create a new #raptor_www object over an existing WWW connection.
+ *
+ * At present this only works with a libcurl CURL handle object
+ * when raptor is compiled with libcurl suppport. Otherwise the
+ * @connection is ignored.  This allows such things as setting
+ * up special flags on the curl handle before passing into the constructor.
  * 
  * Return value: a new #raptor_www object or NULL on failure.
  **/
 raptor_www* 
-raptor_www_new_with_connection(void *connection)
+raptor_www_new_with_connection_v2(raptor_world* world, void *connection)
 {
   raptor_www* www=(raptor_www* )RAPTOR_CALLOC(www, 1, sizeof(raptor_www));
   if(!www)
     return NULL;
-  
+
+  www->world=world;  
   www->type=NULL;
   www->free_type=1; /* default is to free content type */
   www->total_bytes=0;
@@ -150,6 +283,8 @@ raptor_www_new_with_connection(void *connection)
   www->write_bytes=NULL;
   www->content_type=NULL;
   www->uri_filter=NULL;
+  www->connection_timeout=10;
+  www->cache_control=NULL;
 
 #ifdef RAPTOR_WWW_LIBCURL
   www->curl_handle=(CURL*)connection;
@@ -162,21 +297,44 @@ raptor_www_new_with_connection(void *connection)
   raptor_www_libfetch_init(www);
 #endif
 
+  www->error_handlers.locator=&www->locator;
+  raptor_error_handlers_init_v2(world, &www->error_handlers);
+
   return www;
 }
 
 
+#ifndef RAPTOR_DISABLE_V1
 /**
  * raptor_www_new:
  * 
  * Constructor - create a new #raptor_www object.
+ *
+ * raptor_init() MUST have been called before calling this function.
+ * Use raptor_www_new_v2() if using raptor_world APIs.
  * 
  * Return value: a new #raptor_www or NULL on failure.
  **/
 raptor_www*
 raptor_www_new(void)
 {
-  return raptor_www_new_with_connection(NULL);
+  return raptor_www_new_v2(raptor_world_instance());
+}
+#endif
+
+
+/**
+ * raptor_www_new_v2:
+ * @world: raptor_world object
+ * 
+ * Constructor - create a new #raptor_www object.
+ * 
+ * Return value: a new #raptor_www or NULL on failure.
+ **/
+raptor_www*
+raptor_www_new_v2(raptor_world* world)
+{
+  return raptor_www_new_with_connection_v2(world, NULL);
 }
 
 
@@ -201,6 +359,11 @@ raptor_www_free(raptor_www* www)
     www->user_agent=NULL;
   }
 
+  if(www->cache_control) {
+    RAPTOR_FREE(cstring, www->cache_control);
+    www->cache_control=NULL;
+  }
+
   if(www->proxy) {
     RAPTOR_FREE(cstring, www->proxy);
     www->proxy=NULL;
@@ -222,7 +385,10 @@ raptor_www_free(raptor_www* www)
 #endif
 
   if(www->uri)
-    raptor_free_uri(www->uri);
+    raptor_free_uri_v2(www->world, www->uri);
+
+  if(www->final_uri)
+    raptor_free_uri_v2(www->world, www->final_uri);
 
   RAPTOR_FREE(www, www);
 }
@@ -242,11 +408,11 @@ raptor_www_free(raptor_www* www)
  **/
 void
 raptor_www_set_error_handler(raptor_www* www, 
-                             raptor_message_handler error_handler, 
+                             raptor_message_handler error_handler,
                              void *error_data)
 {
-  www->error_handler=error_handler;
-  www->error_data=error_data;
+  www->error_handlers.handlers[RAPTOR_LOG_LEVEL_ERROR].user_data=error_data;
+  www->error_handlers.handlers[RAPTOR_LOG_LEVEL_ERROR].handler=error_handler;
 }
 
 
@@ -301,7 +467,14 @@ raptor_www_set_content_type_handler(raptor_www* www,
 void
 raptor_www_set_user_agent(raptor_www* www, const char *user_agent)
 {
-  char *ua_copy=(char*)RAPTOR_MALLOC(cstring, strlen(user_agent)+1);
+  char *ua_copy=NULL;
+
+  if(!user_agent || !*user_agent) {
+    www->user_agent=NULL;
+    return;
+  }
+  
+  ua_copy=(char*)RAPTOR_MALLOC(cstring, strlen(user_agent)+1);
   if(!ua_copy)
     return;
   strcpy(ua_copy, user_agent);
@@ -322,7 +495,12 @@ raptor_www_set_user_agent(raptor_www* www, const char *user_agent)
 void
 raptor_www_set_proxy(raptor_www* www, const char *proxy)
 {
-  char *proxy_copy=(char*)RAPTOR_MALLOC(cstring, strlen(proxy)+1);
+  char *proxy_copy;
+
+  if(!proxy)
+    return;
+  
+  proxy_copy=(char*)RAPTOR_MALLOC(cstring, strlen(proxy)+1);
   if(!proxy_copy)
     return;
   strcpy(proxy_copy, proxy);
@@ -366,6 +544,73 @@ raptor_www_set_http_accept(raptor_www* www, const char *value)
 }
 
 
+/**
+ * raptor_www_set_connection_timeout:
+ * @www: WWW object
+ * @timeout: Timeout in seconds
+ * 
+ * Set WWW connection timeout
+ **/
+void
+raptor_www_set_connection_timeout(raptor_www* www, int timeout)
+{
+  www->connection_timeout=timeout;
+}
+
+
+/**
+ * raptor_www_set_http_cache_control:
+ * @www: WWW object
+ * @cache_control: Cache-Control header value (or NULL to disable)
+ *
+ * Set HTTP Cache-Control:header (default none)
+ *
+ * The @cache_control value can be a string to set it, "" to send
+ * a blank header or NULL to not set the header at all.
+ *
+ * Return value: non-0 on failure
+ **/
+int
+raptor_www_set_http_cache_control(raptor_www* www, const char* cache_control)
+{
+  char *cache_control_copy;
+  const char* const header="Cache-Control:";
+  const size_t header_len=14; /* strlen("Cache-Control:") */
+  size_t len;
+  
+  RAPTOR_ASSERT((strlen(header) != header_len), "Cache-Control header length is wrong");
+
+  if(www->cache_control)
+    RAPTOR_FREE(cstring, www->cache_control);
+
+  if(!cache_control) {
+    www->cache_control=NULL;
+    return 0;
+  }
+  
+  len=header_len + 1 +strlen(cache_control) + 1; /* header+" "+cache_control+"\0" */
+  
+  cache_control_copy=(char*)RAPTOR_MALLOC(cstring, len);
+  if(!cache_control_copy)
+    return 1;
+  
+  www->cache_control=cache_control_copy;
+
+  strncpy(cache_control_copy, header, header_len);
+  cache_control_copy+= header_len;
+  if(*cache_control) {
+    *cache_control_copy++=' ';
+    strcpy(cache_control_copy, cache_control);
+  }
+  
+#if RAPTOR_DEBUG > 1
+  RAPTOR_DEBUG2("Using Cache-Control header: '%s'\n", www->cache_control);
+#endif
+
+  return 0;
+}
+
+
 /**
  * raptor_www_set_uri_filter:
  * @www: WWW object
@@ -437,34 +682,20 @@ raptor_www_abort(raptor_www* www, const char *reason)
 }
 
 
-void
-raptor_www_error_varargs(raptor_www* www, const char *message, 
-                         va_list arguments)
-{
-  if(www->error_handler) {
-    char *buffer=raptor_vsnprintf(message, arguments);
-    if(!buffer) {
-      fprintf(stderr, "raptor_www_error: Out of memory\n");
-      return;
-    }
-    www->error_handler(www->error_data, &www->locator, buffer);
-    RAPTOR_FREE(cstring, buffer);
-  } else {
-    raptor_print_locator(stderr, &www->locator);
-    fprintf(stderr, " raptor www error - ");
-    vfprintf(stderr, message, arguments);
-    fputc('\n', stderr);
-  }
-}
-
-  
 void
 raptor_www_error(raptor_www* www, const char *message, ...) 
 {
   va_list arguments;
 
   va_start(arguments, message);
-  raptor_www_error_varargs(www, message, arguments);
+
+  raptor_log_error_varargs(www->world,
+                           RAPTOR_LOG_LEVEL_ERROR,
+                           www->error_handlers.handlers[RAPTOR_LOG_LEVEL_ERROR].handler,
+                           www->error_handlers.handlers[RAPTOR_LOG_LEVEL_ERROR].user_data,
+                           &www->locator,
+                           message, arguments);
+
   va_end(arguments);
 }
 
@@ -472,13 +703,14 @@ raptor_www_error(raptor_www* www, const char *message, ...)
 static int 
 raptor_www_file_handle_fetch(raptor_www* www, FILE* fh) 
 {
-  unsigned char buffer[RAPTOR_WWW_BUFFER_SIZE];
+  unsigned char buffer[RAPTOR_WWW_BUFFER_SIZE+1];
   
   while(!feof(fh)) {
     int len=fread(buffer, 1, RAPTOR_WWW_BUFFER_SIZE, fh);
     if(len > 0) {
       www->total_bytes += len;
-
+      buffer[len]='\0';
+      
       if(www->write_bytes)
         www->write_bytes(www, www->write_bytes_userdata, buffer, len, 1);
     }
@@ -499,7 +731,7 @@ raptor_www_file_fetch(raptor_www* www)
 {
   char *filename;
   FILE *fh;
-  unsigned char *uri_string=raptor_uri_as_string(www->uri);
+  unsigned char *uri_string=raptor_uri_as_string_v2(www->world, www->uri);
 #if defined(HAVE_UNISTD_H) && defined(HAVE_SYS_STAT_H)
   struct stat buf;
 #endif
@@ -555,7 +787,7 @@ raptor_www_fetch(raptor_www *www, raptor_uri *uri)
 {
   int status=1;
   
-  www->uri=raptor_new_uri_for_retrieval(uri);
+  www->uri=raptor_new_uri_for_retrieval_v2(www->world, uri);
   
   www->locator.uri=uri;
   www->locator.line= -1;
@@ -569,7 +801,7 @@ raptor_www_fetch(raptor_www *www, raptor_uri *uri)
   status=raptor_www_file_fetch(www);
 #else
 
-  if(raptor_uri_uri_string_is_file_uri(raptor_uri_as_string(www->uri)))
+  if(raptor_uri_uri_string_is_file_uri(raptor_uri_as_string_v2(www->world, www->uri)))
     status=raptor_www_file_fetch(www);
   else {
 #ifdef RAPTOR_WWW_LIBCURL
@@ -586,8 +818,11 @@ raptor_www_fetch(raptor_www *www, raptor_uri *uri)
   }
   
 #endif
-  if(!status && www->status_code && www->status_code != 200)
+  if(!status && www->status_code && www->status_code != 200){
+    raptor_www_error(www, "Resolving URI failed with HTTP status %d",
+                     www->status_code);
     status=1;
+  }
 
   www->failed=status;
   
@@ -623,6 +858,7 @@ raptor_www_fetch_to_string_write_bytes(raptor_www* www, void *userdata,
  * 
  * Return value: non-0 on failure
  **/
+RAPTOR_EXTERN_C
 int
 raptor_www_fetch_to_string(raptor_www *www, raptor_uri *uri,
                            void **string_p, size_t *length_p,
@@ -666,3 +902,42 @@ raptor_www_fetch_to_string(raptor_www *www, raptor_uri *uri,
 
   return (str == NULL);
 }
+
+
+/**
+ * raptor_www_get_final_uri:
+ * @www: #raptor_www object 
+ *
+ * Get the WWW final resolved URI.
+ * 
+ * This returns the URI used after any protocol redirection.
+ *
+ * Return value: a new URI or NULL if not known.
+ **/
+raptor_uri*
+raptor_www_get_final_uri(raptor_www* www) 
+{
+  return www->final_uri ? raptor_uri_copy_v2(www->world, www->final_uri) : NULL;
+}
+
+
+/**
+ * raptor_www_set_final_uri_handler:
+ * @www: WWW object
+ * @handler: content type handler function
+ * @user_data: content type handler data
+ * 
+ * Set the handler to receive the HTTP Content-Type header value.
+ *
+ * This is called if or when the value is discovered during retrieval
+ * by the raptor_www implementation.  Not all implementations provide
+ * access to this.
+ **/
+void
+raptor_www_set_final_uri_handler(raptor_www* www, 
+                                 raptor_www_final_uri_handler handler, 
+                                 void *user_data)
+{
+  www->final_uri_handler=handler;
+  www->final_uri_userdata=user_data;
+}
