@@ -2,12 +2,8 @@
  *
  * rdfdiff.c - Raptor RDF diff tool
  *
- * $Id: rdfdiff.c,v 1.8 2005/09/21 22:57:40 dajobe Exp $
- *
- * Copyright (C) 2000-2005, David Beckett http://purl.org/net/dajobe/
- * Institute for Learning and Research Technology http://www.ilrt.bristol.ac.uk/
- * University of Bristol, UK http://www.bristol.ac.uk/
- *
+ * Copyright (C) 2000-2008, David Beckett http://www.dajobe.org/
+ * Copyright (C) 2000-2005, University of Bristol, UK http://www.bristol.ac.uk/
  * Copyright (C) 2005, Steve Shepard steveshep@gmail.com
  * 
  * This package is Free Software and part of Redland http://librdf.org/
@@ -67,16 +63,17 @@ extern char *optarg;
 #define RDF_NAMESPACE_URI_LEN 43
 #define ORDINAL_STRING_LEN (RDF_NAMESPACE_URI_LEN + MAX_ASCII_INT_SIZE + 1)
 
-#define GETOPT_STRING "bhf:t:"
+#define GETOPT_STRING "bhf:t:u:"
 
 #ifdef HAVE_GETOPT_LONG
-static struct option long_options[] =
+static const struct option long_options[] =
 {
   /* name, has_arg, flag, val */
   {"brief"       , 0, 0, 'b'},
   {"help"        , 0, 0, 'h'},
   {"from-format" , 1, 0, 'f'},
   {"to-format"   , 1, 0, 't'},
+  {"base-uri"    , 1, 0, 'u'},
   {NULL          , 0, 0, 0}
 };
 #endif
@@ -98,6 +95,7 @@ typedef struct rdfdiff_link_s {
 
 typedef struct rdfdiff_blank_s {
   struct rdfdiff_blank_s *next;
+  raptor_world *world;
   char *blank_id;
   raptor_statement *owner;
   rdfdiff_link *first;
@@ -106,6 +104,7 @@ typedef struct rdfdiff_blank_s {
 } rdfdiff_blank;
 
 typedef struct {
+  raptor_world *world;
   char *name;
   raptor_parser *parser;
   rdfdiff_link *first;
@@ -120,20 +119,20 @@ typedef struct {
 
 static int brief = 0;
 static char *program=NULL;
-static const char *title_format_string="Raptor RDF diff utility %s\n";
+static const char * const title_format_string="Raptor RDF diff utility %s\n";
 static int ignore_errors = 0;
 static int ignore_warnings = 0;
 static int emit_from_header = 1;
 static int emit_to_header = 1;
 
-static rdfdiff_file*from_file = NULL;
+static rdfdiff_file* from_file = NULL;
 static rdfdiff_file*to_file = NULL;
 
-static rdfdiff_file* rdfdiff_new_file(const unsigned char *name, const char *syntax);
-static void rdfdiff_free_file(rdfdiff_file*file);
+static rdfdiff_file* rdfdiff_new_file(raptor_world* world, const unsigned char *name, const char *syntax);
+static void rdfdiff_free_file(rdfdiff_file* file);
 
 static rdfdiff_blank *rdfdiff_find_blank(rdfdiff_blank *first, char *blank_id);
-static rdfdiff_blank *rdfdiff_new_blank(char *blank_id);
+static rdfdiff_blank *rdfdiff_new_blank(raptor_world *world, char *blank_id);
 static void rdfdiff_free_blank(rdfdiff_blank *blank);
 
 static int  rdfdiff_blank_equals(const rdfdiff_blank *b1, const rdfdiff_blank *b2,
@@ -143,7 +142,6 @@ static void rdfdiff_error_handler(void *data, raptor_locator *locator, const cha
 static void rdfdiff_warning_handler(void *data, raptor_locator *locator, const char *message);
 
 static void rdfdiff_collect_statements(void *user_data, const raptor_statement *statement);
-static void rdfdiff_compare_statements(void *user_data, const raptor_statement *statement);
 
 int main(int argc, char *argv[]);
 
@@ -169,7 +167,7 @@ safe_strcmp(const char *s1, const char *s2)
 
 #ifdef RDFDIFF_DEBUG
 static void
-rdfdiff_print_statements(rdfdiff_file*file)
+rdfdiff_print_statements(rdfdiff_file* file)
 {    
   fprintf(stderr, "Statements in %s\n",  file->name);
   rdfdiff_link *cur = file->first;
@@ -183,14 +181,15 @@ rdfdiff_print_statements(rdfdiff_file*file)
 
 
 static rdfdiff_file*
-rdfdiff_new_file(const unsigned char *name, const char *syntax)
+rdfdiff_new_file(raptor_world *world, const unsigned char *name, const char *syntax)
 {
-  rdfdiff_file*file = (rdfdiff_file*)RAPTOR_CALLOC(rdfdiff_file, 1, sizeof(rdfdiff_file));
+  rdfdiff_file* file = (rdfdiff_file*)RAPTOR_CALLOC(rdfdiff_file, 1, sizeof(rdfdiff_file));
   if(file) {
+    file->world = world;
     file->name = (char*)RAPTOR_MALLOC(cstring, strlen((const char*)name)+1);
     strcpy((char*)file->name, (const char*)name);
     
-    file->parser = raptor_new_parser(syntax);
+    file->parser = raptor_new_parser_v2(world, syntax);
     if(file->parser) {
       raptor_set_error_handler(file->parser, file, rdfdiff_error_handler);
       raptor_set_warning_handler(file->parser, file, rdfdiff_warning_handler);
@@ -223,7 +222,7 @@ rdfdiff_free_file(rdfdiff_file* file)
   for(cur = file->first; cur; cur = next) {
     next = cur->next;
 
-    raptor_free_statement(cur->statement);
+    raptor_free_statement(file->world, cur->statement);
     RAPTOR_FREE(rdfdiff_link, cur);
   }
 
@@ -239,11 +238,12 @@ rdfdiff_free_file(rdfdiff_file* file)
 
 
 static rdfdiff_blank *
-rdfdiff_new_blank(char *blank_id) 
+rdfdiff_new_blank(raptor_world* world, char *blank_id) 
 {
   rdfdiff_blank *blank = (rdfdiff_blank *)RAPTOR_CALLOC(rdfdiff_blank, 1, sizeof(rdfdiff_blank));
 
   if(blank) {
+    blank->world = world;
     blank->blank_id = (char*)RAPTOR_MALLOC(cstring, strlen(blank_id)+1);
     strcpy((char*)blank->blank_id, (const char*)blank_id);
   }
@@ -261,12 +261,12 @@ rdfdiff_free_blank(rdfdiff_blank *blank)
     RAPTOR_FREE(cstring, blank->blank_id);
 
   if(blank->owner)
-    raptor_free_statement(blank->owner);
+    raptor_free_statement(blank->world, blank->owner);
   
   for(cur = blank->first; cur; cur = next) {
     next = cur->next;
 
-    raptor_free_statement(cur->statement);
+    raptor_free_statement(blank->world, cur->statement);
     RAPTOR_FREE(rdfdiff_link, cur);
   }
 
@@ -276,7 +276,7 @@ rdfdiff_free_blank(rdfdiff_blank *blank)
 
 
 static int
-rdfdiff_ordinal_equals_resource(int ordinal, raptor_uri *resource) 
+rdfdiff_ordinal_equals_resource(raptor_world* world, int ordinal, raptor_uri *resource) 
 {
   unsigned char ordinal_string[ORDINAL_STRING_LEN + 1];
   raptor_uri *ordinal_uri;
@@ -285,43 +285,59 @@ rdfdiff_ordinal_equals_resource(int ordinal, raptor_uri *resource)
   snprintf((char *)ordinal_string, ORDINAL_STRING_LEN, "%s_%d",
            raptor_rdf_namespace_uri, ordinal);
   
-  ordinal_uri = raptor_new_uri(ordinal_string);
+  ordinal_uri = raptor_new_uri_v2(world, ordinal_string);
 
-  equal = raptor_uri_equals(ordinal_uri, resource);
+  equal = raptor_uri_equals_v2(world, ordinal_uri, resource);
     
-  raptor_free_uri(ordinal_uri);
+  raptor_free_uri_v2(world, ordinal_uri);
 
   return equal;
 }
 
 
 static int
-rdfdiff_statement_equals(const raptor_statement *s1, const raptor_statement *s2)
+rdfdiff_statement_equals(raptor_world *world, const raptor_statement *s1, const raptor_statement *s2)
 {
+  int rv=0;
   
   if(!s1 || !s2)
     return 0;
 
+#if RAPTOR_DEBUG > 2
+  fprintf(stderr, "(rdfdiff_statement_equals) Comparing ");
+  raptor_print_statement(s1, stderr);
+  fprintf(stderr, " to ");
+  raptor_print_statement(s2, stderr);
+#endif
+
   if(s1->subject_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL &&
      s2->subject_type == RAPTOR_IDENTIFIER_TYPE_RESOURCE) {
 
     /* check for ordinal/resource equivalence */
-    if(!rdfdiff_ordinal_equals_resource(*(int *)s1->subject, 
-                                        (raptor_uri *)s2->subject))
-      return 0;
+    if(!rdfdiff_ordinal_equals_resource(world,
+                                        *(int *)s1->subject, 
+                                        (raptor_uri *)s2->subject)) {
+      rv=0;
+      goto done;
+    }
     
   } else if(s1->subject_type == RAPTOR_IDENTIFIER_TYPE_RESOURCE &&
             s2->subject_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL) {
 
     /* check for ordinal/resource equivalence */
-    if(!rdfdiff_ordinal_equals_resource(*(int *)s2->subject,
-                                        (raptor_uri *)s1->subject))
-      return 0;
+    if(!rdfdiff_ordinal_equals_resource(world,
+                                        *(int *)s2->subject,
+                                        (raptor_uri *)s1->subject)) {
+      rv=0;
+      goto done;
+    }
       
   } else {
     /* normal comparison */
-    if(s1->subject_type != s2->subject_type)
-      return 0;
+    if(s1->subject_type != s2->subject_type) {
+      rv=0;
+      goto done;
+    }
 
     if(s1->subject_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
       /* Here for completeness. Anonymous nodes are taken care of
@@ -329,45 +345,63 @@ rdfdiff_statement_equals(const raptor_statement *s1, const raptor_statement *s2)
       /*if(strcmp((const char *)s1->subject, (const char *)s2->subject) != 0)
         return 0;*/
     } else {
-      if(!raptor_uri_equals((raptor_uri *)s1->subject,
-                            (raptor_uri *)s2->subject))
-        return 0;
+      if(!raptor_uri_equals_v2(world,
+                               (raptor_uri *)s1->subject,
+                               (raptor_uri *)s2->subject)) {
+        rv=0;
+        goto done;
+      }
     }
   }
 
   if(s1->predicate_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL &&
-      s2->predicate_type == RAPTOR_IDENTIFIER_TYPE_PREDICATE) {
+     s2->predicate_type == RAPTOR_IDENTIFIER_TYPE_PREDICATE) {
 
     /* check for ordinal/resource equivalence */
-    if(!rdfdiff_ordinal_equals_resource(*(int *)s1->predicate, 
-                                        (raptor_uri *)s2->predicate))
-      return 0;
+    if(!rdfdiff_ordinal_equals_resource(world,
+                                        *(int *)s1->predicate, 
+                                        (raptor_uri *)s2->predicate)) {
+      rv=0;
+      goto done;
+    }
 
   } else if(s1->predicate_type == RAPTOR_IDENTIFIER_TYPE_PREDICATE &&
             s2->predicate_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL) {
 
     /* check for ordinal/resource equivalence */
-    if(!rdfdiff_ordinal_equals_resource(*(int *)s2->predicate,
-                                        (raptor_uri *)s1->predicate))
-      return 0;
+    if(!rdfdiff_ordinal_equals_resource(world,
+                                        *(int *)s2->predicate,
+                                        (raptor_uri *)s1->predicate)) {
+      rv=0;
+      goto done;
+    }
       
   } else {
     
-    if(s1->predicate_type != s2->predicate_type)
-      return 0;
+    if(s1->predicate_type != s2->predicate_type) {
+      rv=0;
+      goto done;
+    }
   
     if(s1->predicate_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL) {
-      if(*(int *)s1->predicate != *(int *)s2->predicate)
-        return 0;
+      if(*(int *)s1->predicate != *(int *)s2->predicate) {
+        rv=0;
+        goto done;
+      }
     } else {
-      if(!raptor_uri_equals((raptor_uri *)s1->predicate,
-                            (raptor_uri *)s2->predicate))
-        return 0;
+      if(!raptor_uri_equals_v2(world,
+                               (raptor_uri *)s1->predicate,
+                               (raptor_uri *)s2->predicate)) {
+        rv=0;
+        goto done;
+      }
     }
   }
   
-  if(s1->object_type != s2->object_type)
-    return 0;
+  if(s1->object_type != s2->object_type) {
+    rv=0;
+    goto done;
+  }
   
   if(s1->object_type == RAPTOR_IDENTIFIER_TYPE_LITERAL || 
      s1->object_type == RAPTOR_IDENTIFIER_TYPE_XML_LITERAL) {
@@ -385,25 +419,35 @@ rdfdiff_statement_equals(const raptor_statement *s1, const raptor_statement *s2)
         equal=1;
 
       if(equal)
-        equal=raptor_uri_equals(s1->object_literal_datatype, 
-                              s2->object_literal_datatype);
+        equal=raptor_uri_equals_v2(world,
+                                   s1->object_literal_datatype, 
+                                   s2->object_literal_datatype);
     }
 
-    return equal;
+    rv=equal;
+    goto done;
   } else if(s1->object_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
     /* Here for completeness. Anonymous nodes are taken care of
      * elsewhere */
     /* if(strcmp((const char *)s1->object, (const char *)s2->object) != 0)
        return 0; */
   } else if(s1->object_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL) {
-    if(*(int *)s1->object != *(int *)s2->object)
-      return 0;
+    if(*(int *)s1->object != *(int *)s2->object) {
+      rv=0;
+      goto done;
+    }
   } else {
-    if(!raptor_uri_equals((raptor_uri *)s1->object, (raptor_uri *)s2->object))
-      return 0;
+    if(!raptor_uri_equals_v2(world, (raptor_uri *)s1->object, (raptor_uri *)s2->object))
+      rv=0;
   }
 
-  return 1;
+  rv=1;
+  done:
+
+#if RAPTOR_DEBUG > 2
+  fprintf(stderr, " : %s\n", (rv ? "equal" : "not equal"));
+#endif
+  return rv;
 }
 
 
@@ -427,7 +471,7 @@ rdfdiff_blank_equals(const rdfdiff_blank *b1, const rdfdiff_blank *b2,
     /* Neither are anonymous. Normal comparison. This will return
      * false if both the subject and the predicates don't match. We
      * know the objects are blank nodes. */
-    equal = rdfdiff_statement_equals(b1->owner, b2->owner);
+    equal = rdfdiff_statement_equals(b1->world, b1->owner, b2->owner);
     
   } else if(b1->owner->subject_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS &&
              b2->owner->subject_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
@@ -462,7 +506,7 @@ rdfdiff_blank_equals(const rdfdiff_blank *b1, const rdfdiff_blank *b2,
       rdfdiff_link *s2 = b2->first;
       while (s2) {
 
-        if(rdfdiff_statement_equals(s1->statement, s2->statement))
+        if(rdfdiff_statement_equals(b1->world, s1->statement, s2->statement))
           break;
           
         s2 = s2->next;
@@ -488,11 +532,11 @@ static void
 rdfdiff_error_handler(void *data, raptor_locator *locator,
                       const char *message)
 {
-  rdfdiff_file*file = (rdfdiff_file*)data;
+  rdfdiff_file* file = (rdfdiff_file*)data;
   
   if(!ignore_errors) {
     fprintf(stderr, "%s: Error - ", program);
-    raptor_print_locator(stderr, locator);
+    raptor_print_locator_v2(file->world, stderr, locator);
     fprintf(stderr, " - %s\n", message);
     
     raptor_parse_abort(file->parser);
@@ -507,11 +551,11 @@ static void
 rdfdiff_warning_handler(void *data, raptor_locator *locator,
                         const char *message) 
 {
-  rdfdiff_file*file = (rdfdiff_file*)data;
+  rdfdiff_file* file = (rdfdiff_file*)data;
 
   if(!ignore_warnings) {
     fprintf(stderr, "%s: Warning - ", program);
-    raptor_print_locator(stderr, locator);
+    raptor_print_locator_v2(file->world, stderr, locator);
     fprintf(stderr, " - %s\n", message);
   }
 
@@ -543,12 +587,12 @@ rdfdiff_find_blank(rdfdiff_blank *first, char *blank_id)
 
 
 static rdfdiff_blank *
-rdfdiff_lookup_blank(rdfdiff_file*file, char *blank_id) 
+rdfdiff_lookup_blank(rdfdiff_file* file, char *blank_id) 
 {
   rdfdiff_blank *rv_blank = rdfdiff_find_blank(file->first_blank, blank_id);
   
   if(rv_blank == NULL) {
-    rv_blank = rdfdiff_new_blank(blank_id);
+    rv_blank = rdfdiff_new_blank(file->world, blank_id);
     if(rv_blank) {
 
       if(!file->first_blank) {
@@ -567,7 +611,7 @@ rdfdiff_lookup_blank(rdfdiff_file*file, char *blank_id)
 
 
 static int
-rdfdiff_add_blank_statement(rdfdiff_file*file, const raptor_statement *statement)
+rdfdiff_add_blank_statement(rdfdiff_file* file, const raptor_statement *statement)
 {
   int rv = 0;
   
@@ -578,7 +622,7 @@ rdfdiff_add_blank_statement(rdfdiff_file*file, const raptor_statement *statement
 
     if(dlink) {
 
-      dlink->statement = raptor_statement_copy(statement);
+      dlink->statement = raptor_statement_copy(file->world, statement);
 
       if(dlink->statement) {
     
@@ -614,13 +658,13 @@ rdfdiff_add_blank_statement(rdfdiff_file*file, const raptor_statement *statement
 
 
 static int
-rdfdiff_add_blank_statement_owner(rdfdiff_file*file, const raptor_statement *statement)
+rdfdiff_add_blank_statement_owner(rdfdiff_file* file, const raptor_statement *statement)
 {
   int rv = 0;
   
   rdfdiff_blank *blank = rdfdiff_lookup_blank(file, (char *)statement->object);
   if(blank) {
-    blank->owner = raptor_statement_copy(statement);
+    blank->owner = raptor_statement_copy(file->world, statement);
 
     if(!blank->owner)
       rv = 1;
@@ -636,7 +680,7 @@ rdfdiff_add_blank_statement_owner(rdfdiff_file*file, const raptor_statement *sta
 
 
 static int
-rdfdiff_add_statement(rdfdiff_file*file, const raptor_statement *statement) 
+rdfdiff_add_statement(rdfdiff_file* file, const raptor_statement *statement) 
 {
   int rv = 0;
   
@@ -644,7 +688,7 @@ rdfdiff_add_statement(rdfdiff_file*file, const raptor_statement *statement)
 
   if(dlink) {
 
-    dlink->statement = raptor_statement_copy(statement);
+    dlink->statement = raptor_statement_copy(file->world, statement);
 
     if(dlink->statement) {
       
@@ -675,6 +719,37 @@ rdfdiff_add_statement(rdfdiff_file*file, const raptor_statement *statement)
 }
 
 
+static rdfdiff_link*
+rdfdiff_statement_find(rdfdiff_file* file, const raptor_statement *statement,
+                       rdfdiff_link** prev_p)
+{
+  rdfdiff_link* prev = NULL;
+  rdfdiff_link* cur = file->first;
+  
+  while(cur) {
+    if(rdfdiff_statement_equals(file->world, cur->statement, statement)) {
+      if(prev_p)
+        *prev_p=prev;
+      return cur;
+    }
+    prev=cur;
+    cur=cur->next;
+  }
+
+  return NULL;
+}
+
+
+static int
+rdfdiff_statement_exists(rdfdiff_file* file, const raptor_statement *statement)
+{
+  rdfdiff_link* node;
+  rdfdiff_link* prev=NULL;
+  node=rdfdiff_statement_find(file, statement, &prev);
+  return (node != NULL);
+}
+
+
 /*
  * rdfdiff_collect_statements - Called when parsing "from" file to build a
  * list of statements for comparison with those in the "to" file.
@@ -683,8 +758,11 @@ static void
 rdfdiff_collect_statements(void *user_data, const raptor_statement *statement)
 {
   int rv = 0;
+  rdfdiff_file* file = (rdfdiff_file*)user_data;
+
+  if(rdfdiff_statement_exists(file, statement))
+    return;
   
-  rdfdiff_file*file = (rdfdiff_file*)user_data;
   file->statement_count++;
 
   if(statement->subject_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS ||
@@ -707,91 +785,18 @@ rdfdiff_collect_statements(void *user_data, const raptor_statement *statement)
 }
 
 
-/*
- * rdfdiff_compare_statements - Called when parsing "to" file to compare the
- * statements with those in the "from" file.
- */
-static void
-rdfdiff_compare_statements(void *user_data, const raptor_statement *statement)
-{
-  int rv = 0;
-  
-  rdfdiff_file*file = (rdfdiff_file*)user_data; /* file == &to_file */
-  file->statement_count++;
-
-  if(statement->subject_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS ||
-      statement->object_type  == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
-
-    /* We don't have enough information yet to make the
-     * comparison. We'll have to wait until the parsing is complete.
-     */
-    if(statement->subject_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS)
-      rv = rdfdiff_add_blank_statement(file, statement);
-
-    if(rv == 0 && statement->object_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS)
-      rv = rdfdiff_add_blank_statement_owner(file, statement);
-
-    if(rv != 0)
-      raptor_parse_abort(file->parser);
-
-  } else {
-
-    rdfdiff_link *prev = 0;
-    rdfdiff_link *cur = from_file->first;
-
-    while (cur) {
-
-      if(rdfdiff_statement_equals(cur->statement, statement)) {
-        /* remove it from the list */
-        if(from_file->first == cur) {
-          from_file->first = cur->next;
-        } else {
-          prev->next = cur->next;
-        }
-
-        raptor_free_statement(cur->statement);
-      
-        RAPTOR_FREE(rdfdiff_link, cur);
-
-        break;
-      
-      } else {
-        prev = cur;
-      }
-    
-      cur = cur->next;
-    
-    }
-
-    if(!cur) {
-
-      if(!brief) {
-
-        if(emit_from_header) {
-          fprintf(stderr, "Statements in %s but not in %s\n",  to_file->name, from_file->name);
-          emit_from_header = 0;
-        }
-    
-        fprintf(stderr, "<    ");
-        raptor_print_statement(statement, stderr);
-        fprintf(stderr, "\n");
-      }
-        
-      file->difference_count++;
-    } 
-  }
-  
-}
 
 int
 main(int argc, char *argv[]) 
 {
+  raptor_world *world = NULL;
   unsigned char *from_string=NULL;
   unsigned char *to_string=NULL;
   raptor_uri *from_uri=NULL;
   raptor_uri *to_uri=NULL;
-  char *from_syntax = "rdfxml";
-  char *to_syntax = "rdfxml";
+  raptor_uri *base_uri=NULL;
+  const char *from_syntax = "rdfxml";
+  const char *to_syntax = "rdfxml";
   int free_from_string = 0;
   int free_to_string = 0;
   int usage=0;
@@ -799,7 +804,8 @@ main(int argc, char *argv[])
   char *p;
   int rv = 0;
   rdfdiff_blank *b1;
-
+  rdfdiff_link *cur;
+  
   program=argv[0];
   if((p=strrchr(program, '/')))
     program=p+1;
@@ -807,7 +813,12 @@ main(int argc, char *argv[])
     program=p+1;
   argv[0]=program;
 
-  raptor_init();
+  world = raptor_new_world();
+  if(!world)
+    exit(1);
+  rv = raptor_world_open(world);
+  if(rv)
+    exit(1);
 
   while (!usage && !help)
   {
@@ -846,6 +857,11 @@ main(int argc, char *argv[])
           to_syntax = optarg;
         break;
 
+      case 'u':
+        if(optarg)
+          base_uri = raptor_new_uri_v2(world, (const unsigned char*)optarg);
+        break;
+
     }
     
   }
@@ -873,8 +889,9 @@ main(int argc, char *argv[])
     puts("\nOPTIONS:");
     puts(HELP_TEXT("h", "help                      ", "Print this help, then exit"));
     puts(HELP_TEXT("b", "brief                     ", "Report only whether files differ"));
-    puts(HELP_TEXT("f FORMAT", "from-format FORMAT ", "Format of <from URI> (default is rdfxml)"));
-    puts(HELP_TEXT("t FORMAT", "to-format FORMAT   ", "Format of <to URI> (default is rdfxml)"));
+    puts(HELP_TEXT("u BASE-URI", "base-uri BASE-URI  ", "Set the base URI for the files"));
+    puts(HELP_TEXT("f FORMAT",   "from-format FORMAT ", "Format of <from URI> (default is rdfxml)"));
+    puts(HELP_TEXT("t FORMAT",   "to-format FORMAT   ", "Format of <to URI> (default is rdfxml)"));
     rv = 1;
     goto exit;
   }
@@ -905,7 +922,7 @@ main(int argc, char *argv[])
   }
   
   if(from_string) {
-    from_uri = raptor_new_uri(from_string);
+    from_uri = raptor_new_uri_v2(world, from_string);
     if(!from_uri) {
       fprintf(stderr, "%s: Failed to create URI for %s\n", program, from_string);
       rv = 2;
@@ -914,7 +931,7 @@ main(int argc, char *argv[])
   }
   
   if(to_string) {
-    to_uri = raptor_new_uri(to_string);
+    to_uri = raptor_new_uri_v2(world, to_string);
     if(!to_uri) {
       fprintf(stderr, "%s: Failed to create URI for %s\n", program, from_string);
       rv = 2;
@@ -923,23 +940,24 @@ main(int argc, char *argv[])
   }
 
   /* create and init "from" data structures */
-  from_file = rdfdiff_new_file(from_string, from_syntax);
+  from_file = rdfdiff_new_file(world, from_string, from_syntax);
   if(!from_file) {
     rv = 2;
     goto exit;
   }
   
   /* create and init "to" data structures */
-  to_file = rdfdiff_new_file(to_string, to_syntax);
+  to_file = rdfdiff_new_file(world, to_string, to_syntax);
   if(!to_file) {
     rv = 2;
     goto exit;
   }
 
   /* parse the files */
-  raptor_set_statement_handler(from_file->parser, from_file, rdfdiff_collect_statements);
+  raptor_set_statement_handler(from_file->parser, from_file,
+                               rdfdiff_collect_statements);
   
-  if(raptor_parse_uri(from_file->parser, from_uri, NULL)) {
+  if(raptor_parse_uri(from_file->parser, from_uri, base_uri)) {
     fprintf(stderr, "%s: Failed to parse URI %s as %s content\n", program, 
             from_string, from_syntax);
     rv = 1;
@@ -947,14 +965,50 @@ main(int argc, char *argv[])
   } else {
 
     /* Note intentional from_uri as base_uri */
-    raptor_set_statement_handler(to_file->parser, to_file, rdfdiff_compare_statements);
-    if(raptor_parse_uri(to_file->parser, to_uri, from_uri)) {
+    raptor_set_statement_handler(to_file->parser, to_file,
+                                 rdfdiff_collect_statements);
+    if(raptor_parse_uri(to_file->parser, to_uri, base_uri ? base_uri: from_uri)) {
       fprintf(stderr, "%s: Failed to parse URI %s as %s content\n", program, 
               to_string, to_syntax);
       rv = 1;
       goto exit;
     }
   }
+
+
+  /* Compare triples with no blank nodes */
+  cur = to_file->first;
+  while(cur) {
+    rdfdiff_link* node;
+    rdfdiff_link* prev;
+    node=rdfdiff_statement_find(from_file, cur->statement, &prev);
+    if(node) {
+      /* exists in from file - remove it from the list */
+      if(from_file->first == node) {
+        from_file->first = node->next;
+      } else {
+        prev->next = node->next;
+      }
+      raptor_free_statement(world, node->statement);
+      RAPTOR_FREE(rdfdiff_link, node);
+    } else {
+      if(!brief) {
+        if(emit_from_header) {
+          fprintf(stderr, "Statements in %s but not in %s\n",
+                  to_file->name, from_file->name);
+          emit_from_header = 0;
+        }
+        
+        fprintf(stderr, "<    ");
+        raptor_print_statement_v1(world, cur->statement, stderr);
+        fprintf(stderr, "\n");
+      }
+      
+      to_file->difference_count++;
+    }
+    cur=cur->next;
+  }
+
   
   /* Now compare the blank nodes */
   b1 = to_file->first_blank;
@@ -1000,10 +1054,10 @@ main(int argc, char *argv[])
   if(from_file->first) {
     /* The entrys left in from_file have not been found in to_file. */
     if(!brief) {
-      rdfdiff_link* cur;
-      
+
       if(emit_to_header) {
-        fprintf(stderr, "Statements in %s but not in %s\n",  from_file->name, to_file->name);
+        fprintf(stderr, "Statements in %s but not in %s\n",  from_file->name,
+                to_file->name);
         emit_to_header = 0;
       }
       
@@ -1011,7 +1065,7 @@ main(int argc, char *argv[])
       while (cur) {
         if(!brief) {
           fprintf(stderr, ">    ");
-          raptor_print_statement(cur->statement, stderr);
+          raptor_print_statement_v1(world, cur->statement, stderr);
           fprintf(stderr, "\n");
         }
       
@@ -1060,6 +1114,9 @@ main(int argc, char *argv[])
 
 exit:
 
+  if(base_uri)
+    raptor_free_uri_v2(world, base_uri);
+  
   if(from_file)
     rdfdiff_free_file(from_file);
 
@@ -1073,12 +1130,12 @@ exit:
     raptor_free_memory(to_string);
 
   if(from_uri)
-    raptor_free_uri(from_uri);
+    raptor_free_uri_v2(world, from_uri);
 
   if(to_uri)
-    raptor_free_uri(to_uri);
+    raptor_free_uri_v2(world, to_uri);
 
-  raptor_finish();
+  raptor_free_world(world);
 
   return rv;
   
